// file: $isip/class/algo/FourierTransform/ft_03.cc
// version: $Id: ft_03.cc,v 1.11 2002/05/31 21:57:29 picone Exp $
//

// isip include files
//
#include "FourierTransform.h"

// ------------------------------------------------------------------
// these methods have to be in the same file so they can use the same
// static parser pointer
// ------------------------------------------------------------------

static SofParser* parser_d = (SofParser*)NULL;

// method: setParser
//
// arguments:
//  SofParser* parser: (input) sof file object
//
// return: a boolean value indicating status
//
// this method sets the parser from the parent object to be used in
// the next readData method
//
boolean FourierTransform::setParser(SofParser* parser_a) {

  // set the parser
  //
  parser_d = parser_a;

  // exit gracefully
  //
  return true;
}

// method: read
//
// arguments:
//  Sof& sof: (input) sof file object
//  long tag: (input) sof object instance tag
//  const String& name: (input) sof object instance name
//
// return: a boolean value indicating status
//
// this method has the object read itself from an Sof file according
// to the specified name and tag
//
boolean FourierTransform::read(Sof& sof_a, long tag_a, const String& name_a) {

  // get the instance of the object from the Sof file
  //
  if (!sof_a.find(name_a, tag_a)) {
    return false;
  }

  // read the actual data from the sof file
  //
  return readData(sof_a);
}

// method: readData
//
// arguments:
//  Sof& sof: (input) sof file object
//  const String& pname: (input) parameter name
//  long size: (input) size in bytes of object (or FULL_SIZE)
//  boolean param: (input) is the parameter specified?
//  boolean nested: (input) is this nested?
//
// return: a boolean value indicating status
//
// this method has the object read itself from an Sof file. it assumes
// that the Sof file is already positioned correctly
//
boolean FourierTransform::readData(Sof& sof_a, const String& pname_a,
				   long size_a, boolean param_a,
				   boolean nested_a) {

  // we need a parser
  // 
  if (parser_d == (SofParser*)NULL) {
    
    parser_d = new SofParser;
    
    if (nested_a) {
      parser_d->setNest();
    }
    
    // load the parser
    //
    if (!parser_d->load(sof_a, size_a)) {

      // delete the parser
      //
      delete parser_d;
      parser_d = (SofParser*)NULL;
      
      // return a warning message
      //
      return Error::handle(name(), L"readData", Error::READ,
                           __FILE__, __LINE__, Error::WARNING);
    }           
  }

  // set the debug level of the parser
  //
  parser_d->setDebug(debug_level_d.getIndex());

  // get the algorithm type
  //
  if (parser_d->isPresent(sof_a, PARAM_ALGORITHM)) {
    if (!ALGO_MAP.readElementData((long&)algorithm_d, sof_a, PARAM_ALGORITHM,
				  parser_d->getEntry(sof_a,
						     PARAM_ALGORITHM))) {

      // delete the parser
      //
      delete parser_d;
      parser_d = (SofParser*)NULL;
      
      // return a warning message
      //
      return Error::handle(name(), L"readData", Error::READ,
			   __FILE__, __LINE__, Error::WARNING);
    }
  }
  else {
    algorithm_d = DEF_ALGORITHM;
  }
  
  // check known algorithms: invoke the proper lower-level method
  //
  if ((algorithm_d == DFT) || (algorithm_d == FFT) || (algorithm_d == DCT)) {
    return readDataCommon(sof_a, pname_a, size_a, param_a, nested_a);
  }

  // check unsupported algorithms
  //
  else {
    return Error::handle(name(), L"readData",
			 ERR_UNKALG, __FILE__, __LINE__);
  }

  return true;
}

// method: readDataCommon
//
// arguments:
//  Sof& sof: (input) sof file object
//  const String& pname: (input) parameter name
//  long size: (input) size in bytes of object (or FULL_OBJECT)
//  boolean param: (input) is the parameter name in the file?
//  boolean nested: (input) are we nested?
//
// return: a boolean value indicating status
//
// this method assumes the algorithm has been read, and proceeds to read
// the rest of the object based on the value of the algorithm
//
boolean FourierTransform::readDataCommon(Sof& sof_a, const String& pname_a,
					 long size_a, boolean param_a,
					 boolean nested_a) {

  // get the implementation
  //
  if (parser_d->isPresent(sof_a, PARAM_IMPLEMENTATION)) {
    if (!IMPL_MAP.readElementData((long&)implementation_d, sof_a,
				  PARAM_IMPLEMENTATION,
				  parser_d->getEntry(sof_a,
						     PARAM_IMPLEMENTATION))) {

      // delete the parser
      //
      delete parser_d;
      parser_d = (SofParser*)NULL;
      
      // return a warning message
      //
      return Error::handle(name(), L"readDataCommon", Error::READ,
			   __FILE__, __LINE__, Error::WARNING);
    }
  }
  else {
    implementation_d = DEF_IMPLEMENTATION;
  }

  // get the direction
  //
  if (parser_d->isPresent(sof_a, PARAM_DIRECTION)) {
    if (!DIRE_MAP.readElementData((long&)direction_d, sof_a,
				  PARAM_DIRECTION,
				  parser_d->getEntry(sof_a,
						     PARAM_DIRECTION))) {

      // delete the parser
      //
      delete parser_d;
      parser_d = (SofParser*)NULL;
      
      // return a warning message
      //
      return Error::handle(name(), L"readDataCommon", Error::READ,
			   __FILE__, __LINE__, Error::WARNING);
    }
  }
  else {
    direction_d = DEF_DIRECTION;
  }

  // get the output mode
  //
  if (parser_d->isPresent(sof_a, PARAM_OTYPE)) {
    if (!OUTM_MAP.readElementData((long&)otype_d, sof_a,
				  PARAM_OTYPE,
				  parser_d->getEntry(sof_a, PARAM_OTYPE))) {

      // delete the parser
      //
      delete parser_d;
      parser_d = (SofParser*)NULL;
      
      // return a warning message
      //
      return Error::handle(name(), L"readDataCommon", Error::READ,
			   __FILE__, __LINE__, Error::WARNING);
    }
  }
  else {
    otype_d = DEF_OTYPE;
  }
  
  // get the resolution
  //
  if (parser_d->isPresent(sof_a, PARAM_RESOLUTION)) {
    if (!RESO_MAP.readElementData((long&)resolution_d, sof_a,
				  PARAM_RESOLUTION,
				  parser_d->getEntry(sof_a,
						     PARAM_RESOLUTION))) {

      // delete the parser
      //
      delete parser_d;
      parser_d = (SofParser*)NULL;
      
      // return a warning message
      //
      return Error::handle(name(), L"readDataCommon", Error::READ,
			   __FILE__, __LINE__, Error::WARNING);
    }
  }
  else {
    resolution_d = DEF_RESOLUTION;
  }

  // get the input length
  //
  if (parser_d->isPresent(sof_a, PARAM_INPUT_LEN)) {
    if (!ilen_d.readData(sof_a, PARAM_INPUT_LEN,
			 parser_d->getEntry(sof_a, PARAM_INPUT_LEN))) {

      // delete the parser
      //
      delete parser_d;
      parser_d = (SofParser*)NULL;

      // return a warning message
      //
      return Error::handle(name(), L"readDataCommon", Error::READ,
			   __FILE__, __LINE__, Error::WARNING);
    }      
    //    fixed_order_d = true;
  }
  else {
    ilen_d = DEF_LENGTH;
  }
  
  // get the output length
  //
  if (parser_d->isPresent(sof_a, PARAM_OUTPUT_LEN)) {
    if (!olen_d.readData(sof_a, PARAM_OUTPUT_LEN,
			 parser_d->getEntry(sof_a, PARAM_OUTPUT_LEN))) {

      // delete the parser
      //
      delete parser_d;
      parser_d = (SofParser*)NULL;

      // return a warning message
      //
      return Error::handle(name(), L"readDataCommon", Error::READ,
			   __FILE__, __LINE__, Error::WARNING);
    }      
    //    fixed_order_d = true;
  }
  else {
    olen_d = DEF_LENGTH;
  }

  // get the FFT order
  //
  if (parser_d->isPresent(sof_a, PARAM_ORDER)) {
    if (!order_d.readData(sof_a, PARAM_ORDER,
			  parser_d->getEntry(sof_a, PARAM_ORDER))) {

      // delete the parser
      //
      delete parser_d;
      parser_d = (SofParser*)NULL;

      // return a warning message
      //
      return Error::handle(name(), L"readDataCommon", Error::READ,
			   __FILE__, __LINE__, Error::WARNING);
    }      
  }
  else {
    order_d = DEF_ORDER;
  }
  
  // get the normalization
  //
  if (parser_d->isPresent(sof_a, PARAM_NORMALIZATION)) {
    if (!NORM_MAP.readElementData((long&)normalization_d, sof_a,
				  PARAM_NORMALIZATION,
				  parser_d->getEntry(sof_a,
						     PARAM_NORMALIZATION))) {

      // delete the parser
      //
      delete parser_d;
      parser_d = (SofParser*)NULL;
      
      // return a warning message
      //
      return Error::handle(name(), L"readDataCommon", Error::READ,
			   __FILE__, __LINE__, Error::WARNING);
    }
  }
  else {
    normalization_d = DEF_NORM;
  }

  // get the debug level
  //
  if (parser_d->isPresent(sof_a, PARAM_DBGL)) {
    if (!debug_level_d.readData(sof_a, PARAM_DBGL,
				parser_d->getEntry(sof_a, PARAM_DBGL))) {
      
      // delete the parser
      //
      delete parser_d;
      parser_d = (SofParser*)NULL;
      
      // return a warning message
      //
      return Error::handle(name(), L"readDataCommon", Error::IO,
			   __FILE__, __LINE__, Error::WARNING);
    }
  }
  else {
    debug_level_d = Integral::DEF_DEBUG;
  }

  // check that all parameters are accounted for
  //
  if (!parser_d->checkParams(sof_a)) {

    // delete the parser
    //
    delete parser_d;
    parser_d = (SofParser*)NULL;

    // return a warning message
    //
    return Error::handle(name(), L"readDataCommon", Error::IO,
			 __FILE__, __LINE__, Error::WARNING);
  }

  // delete the parser
  //
  delete parser_d;
  parser_d = (SofParser*)NULL;

  // exit gracefully
  //
  return true;
}

