// file: $isip/class/algo/Window/win_04.cc
// version: $Id: win_04.cc,v 1.15 2002/05/31 21:57:33 picone Exp $
//

// isip include files
//
#include "Window.h"

// method: sofSize
//
// arguments: none
//
// return: size of object
//
// this method returns the size of the object in the Sof file and is
// used for binary write
//
long Window::sofSize() const {

  // total things common to all algorithms
  //

  // start with the space required for the algorithm name
  //
  long bytes = ALGO_MAP.elementSofSize();

  // add the space required for the implementation name
  //
  bytes += IMPL_MAP.elementSofSize();

  // add the space required for the alignment name
  //
  bytes += ALGN_MAP.elementSofSize();

  // add the space required for the normalization name
  //
  bytes += NORM_MAP.elementSofSize();

  // add the space required for the mode name
  //
  bytes += DMODE_MAP.elementSofSize();
    
  // add the space required for the duration
  //
  bytes += duration_d.sofSize();
  
  // add the space required for the debug level
  //
  bytes += debug_level_d.sofSize();
  
  // total things common to specific algorithms
  //

  // add the space required for window constant vector
  //
  if (algorithm_d != CUSTOM) {
    bytes += constants_d.sofSize();
  }

  // add the space required for the window function
  //
  else {
    bytes += data_d.sofSize();
  }

  // return the size
  //
  return bytes;
}

// method: write
//
// arguments:
//  Sof& sof: (input) sof file object
//  long tag: (input) sof object instance tag
//  const String& name: (input) sof object instance name
//
// return: a boolean value indicating status
//
// this method has the object write itself to an Sof file
//
boolean Window::write(Sof& sof_a, long tag_a, const String& name_a) const {

  // declare local variable
  //
  long obj_size;
  
  // write the instance of the object into the Sof file
  //
  if (sof_a.isText()) {

    // set the size to by dynamic
    //
    obj_size = Sof::ANY_SIZE;
  }
  else {

    // set the size to be the size of the object written to the Sof file
    //
    obj_size = sofSize();
  }

  // write the object into the sof file's index
  //
  if (!sof_a.put(name_a, tag_a, obj_size)) {
    return false;
  }
  
  // write data and exit gracefully
  //
  return writeData(sof_a);
}

// method: writeData
//
// arguments:
//  Sof& sof: (input) sof file object
//  const String& pname: (input) parameter name
//
// return: a boolean value indicating status
//
// this method writes things common to all algorithms, and then
// branches on the algorithm name, and invokes a lower-level method
// to write the algorithm-specific things.
//
boolean Window::writeData(Sof& sof_a, const String& pname_a) const {

  // declare local variables
  //
  boolean status = false;
  
  // write a start string if necessary
  //
  sof_a.writeLabelPrefix(pname_a);

  // write the algorithm name
  //
  ALGO_MAP.writeElementData(sof_a, PARAM_ALGORITHM, (long)algorithm_d);
    
  // write the implementation type
  //
  IMPL_MAP.writeElementData(sof_a, PARAM_IMPLEMENTATION,
  			    (long) implementation_d);
  
  // write the alignment
  //
  ALGN_MAP.writeElementData(sof_a, PARAM_ALIGNMENT, (long)alignment_d);
  
  // write the normalization flag
  //
  NORM_MAP.writeElementData(sof_a, PARAM_NORMALIZATION, (long)normalization_d);

  // write the duration
  //
  duration_d.writeData(sof_a, PARAM_DURATION);

  // write the mode type
  //
  CMODE_MAP.writeElementData(sof_a, PARAM_CMODE, (long)cmode_d);
  
  // write the debug level (from the base class)
  //
  debug_level_d.writeData(sof_a, PARAM_DBGL);

  // check known algorithms: invoke the proper lower-level method
  //  currently, all algorithms use the same i/o methods
  //
  if ((algorithm_d == RECTANGULAR) ||
      (algorithm_d == BARTLETT) ||
      (algorithm_d == BLACKMAN) ||
      (algorithm_d == DOLPH_CHEBYSHEV) ||
      (algorithm_d == GAUSSIAN) ||
      (algorithm_d == HAMMING) ||
      (algorithm_d == HANNING) ||
      (algorithm_d == KAISER) ||
      (algorithm_d == LIFTER) ||
      (algorithm_d == CUSTOM)) {
    return writeDataCommon(sof_a, pname_a);
  }

  // check unknown algorithms
  //
  else {
    return Error::handle(name(), L"writeData",
			 ERR_UNKALG, __FILE__, __LINE__);
  }

  // put an end string if necessary
  //
  sof_a.writeLabelSuffix(pname_a);
  
  // exit gracefully
  //
  return status;
}

// method: writeDataCommon
//
// arguments:
//  Sof& sof: (input) sof file object
//  const String& pname: (input) parameter name
//
// return: a boolean value indicating status
//
// this method writes data to a file for several algorithms that
// share the same parameters
//
boolean Window::writeDataCommon(Sof& sof_a,
				const String& pname_a) const {

  // write constants vector
  //
  if (algorithm_d != CUSTOM) {
    constants_d.writeData(sof_a, PARAM_CONSTANTS);
  }

  // write the data
  //
  else {
    data_d.writeData(sof_a, PARAM_DATA);
  }

  // exit gracefully
  //
  return true;
}
