// file: $isip/class/asr/LanguageModel/lm_04.cc
// version: $Id: lm_11.cc,v 1.19 2002/12/15 23:00:08 parihar Exp $
//

// isip include files
//
#include "LanguageModel.h"

// method: store
//
// arguments:
//  const String& model_file_a: (output) file name for a SOF file containing
//                              language models
//  const String& stat_file_a: (output) file name for a SOF file containing
//                             statistical models associated with the last
//                             level
//  const Vector<SearchLevel>& search_levels_a: (input) a set of search levels
//  OUTPUT_FORMAT format_a: (input) user-defined file format for the output
//                          model file
//  OUTPUT_TYPE type_a: (input) user-defined file type (text/binary)
//
// return: a boolean value indicating status
//
// this method outputs a set of input search levels to two SOF files
// with user-defined filenames according to a user-chosen file format
// that is either NATIVE format or JSGF format
//
boolean LanguageModel::store(const String& model_file_a,
			     const String& stat_file_a,
			     Vector<SearchLevel>& search_levels_a,
			     OUTPUT_FORMAT format_a, OUTPUT_TYPE type_a) {
  
  // get the last level index
  //
  long last_level = search_levels_a.length() - 1;
  boolean context_map = false;
  
  // branch on type: BINARY
  // all format types are included in this case - NATIVE, JSGF
  //
  if (type_a == BINARY) {
    
    // open the input sof file
    //
    Sof model_sof;
    if(!model_sof.open(model_file_a, File::WRITE_ONLY, File::BINARY)) {
      return Error::handle(model_file_a, L"open", Error::TEST,
			   __FILE__, __LINE__);
    }
    
    // store the input search levels into the model file
    //
    for (long i = 0; i < search_levels_a.length(); i++) {
      search_levels_a(i).store(model_sof, i);
    }

    // close the file
    //
    model_sof.close();
  } // end of type = BINARY
  
  // else branch on type: TEXT
  //
  else if (type_a == TEXT) {

    // branch on format: NATIVE
    // this case is same as type: BINARY because the NATIVE format
    // deals with binary format automatically through the Sof class
    // interface
    //
    if (format_a == NATIVE) {
      
      // open the input sof file
      //
      Sof model_sof;
      if(!model_sof.open(model_file_a, File::WRITE_ONLY)) {
	return Error::handle(model_file_a, L"open", Error::TEST,
			     __FILE__, __LINE__);
      }
      
    // store the input search levels into the model file
    //
    for (long i = 0; i < search_levels_a.length(); i++) {
      search_levels_a(i).store(model_sof, i);
    }
    
    // close the file
    //
    model_sof.close();
    } // end of format: NATIVE
    
    // branch on the format: JSGF
    //
    else if (format_a == JSGF) {
      
      // declare a vector of strings to store each level of JSGF grammars
      //
      Vector<String> JSGF_models;
      
      // loop through the search levels
      //
      for (long i = 0; i < search_levels_a.length(); i++) {
	
	// declare a string to store the JSGF grammars in this level
	//
	String JSGF_level;
	context_map = false;
	
	// get the symbol table of the previous search level if available
	//
	long level_index = 0;
	if (i > 0) {
	  level_index = i - 1;
	}
	
	Vector<SearchSymbol>& symbol_table =
	  search_levels_a(level_index).getSymbolTable();
	
	Vector<SearchSymbol> graph_symbols;
	
	// get the number of subgraphs in the level
	//
	long num_graphs = search_levels_a(i).getNumSubGraphs();
	
	// do we have context mapping ?
	//
	if (search_levels_a(level_index).getContextMap().length() > 0){
	  context_map = true;
	  
	  // set the symbol table
	  //
	  for (long j = 0; j < num_graphs; j++) {
	    String symbol;
	    symbol.assign(L"G_");
	    Ulong index;
	    index.assign(j);
	    symbol.concat(index);
	    //symbol.debug(L"symbol");
	    graph_symbols.concat(symbol);
	  }	  
	}
	
	// loop through each subgraph in this level
	//
	for (long j = 0; j < num_graphs; j++) {
	  
	  // added for debug
	  //
	  //Long index(j);
	  //index.debug(L"the index j");
	  
	  // declare a JSGF grammar string
	  //
	  String JSGF_grammar(L"grammar = {\n  #JSGF V1.0;\n");
	  
	  // get the graph
	  //
	  DiGraph<SearchNode>& graph = search_levels_a(i).getSubGraph(j);
	  
	  // get the grammar name for the graph
	  //
	  SearchSymbol grammar_name;
	  
	  // the top level
	  //
	  if (i == 0) {
	    grammar_name.assign(L"sentence");
	  }
	  
	  // for all other level, the graph's grammar name is the corresponding
	  // symbol in the symbol table of the previous level
	  //
	  else {
	    if ( context_map ){
	      grammar_name.assign(graph_symbols(j));	    
	    }
	    else {
	      grammar_name.assign(symbol_table(j));
	    }
	  }
	  
	  // convert the graph to a JSGF grammar string
	  //
	  //JSGF_grammar.concat(convertToJSGF(graph, grammar_name));
	  JSGF_grammar.concat(digraphToJSGF(graph, grammar_name));
	  JSGF_grammar.concat(L"};\n\n");
	  
	  // add this grammar to the level in JSGF format
	  //
	  JSGF_level.concat(JSGF_grammar);	
	} // end: for (long j = 0; j < num_graphs; j++)
	
	// add this level of JSGF to the vector of JSGF models
	//
	JSGF_models.concat(JSGF_level); 
      } // end: for (long i = 0; i < search_levels_a.length(); i++)
      
      // output the JSGF models to the output SOF file
      //
      // declare the line of algorithm tag
      //
      String algo_tag(L"algorithm = \"JSGF\";\n");
      
      // set the size of the object to be written
      //
      long algo_size = algo_tag.length();
      
      // first store in the output SOF file the the grammars of definitions
      // for graph starting and stoping points
      //
      String start_def(L"grammar = {\n  #JSGF V1.0;\n  // Define the grammar name\n  grammar network.grammar.ISIP_JSGF_START;\n\n  // Define the ISIP graph start symbol\n  public <ISIP_JSGF_1_0_START> = S;\n};\n\n");
      
      String term_def(L"grammar = {\n  #JSGF V1.0;\n  // Define the grammar name\n  grammar network.grammar.ISIP_JSGF_TERM;\n\n  // Define the ISIP graph terminal symbol\n  public <ISIP_JSGF_1_0_TERM> = T;\n};\n\n");
      
      // open the input sof file
      //
      Sof model_sof;
      if(!model_sof.open(model_file_a, File::WRITE_ONLY)) {
	return Error::handle(model_file_a, L"open", Error::TEST,
			     __FILE__, __LINE__);
    }
      
      // write the definitions of graph starting and stoping points
      // into the sof file
      //      
      model_sof.put(L"JSGF", 100, algo_size +
		    start_def.length() + term_def.length());
      model_sof.puts(algo_tag);
      model_sof.puts(start_def);
      model_sof.puts(term_def);
      
      // loop through each level and write its JSGF grammars into the SOF file
      //
      for (long i = 0; i < JSGF_models.length(); i++) {
	model_sof.put(L"JSGF", i, algo_size + JSGF_models(i).length());
	model_sof.puts(algo_tag);
	model_sof.puts(JSGF_models(i));
	
	// write other stuff at each level that are not grammar
	//
	storeNonGrammars(model_sof, search_levels_a, i);
      }
      
      // close the file
      //
      model_sof.close();
    } // end of format: JSGF
    
    // else error
    //
    else {
      return Error::handle(name(), L"invalid format", Error::TEST, __FILE__, __LINE__);
    }        
  } // end of type: TEXT
  
  // else error
  //
  else {
    return Error::handle(name(), L"invalid type", Error::TEST, __FILE__, __LINE__);
  }

  // get the statistical model hashtable and the models
  //
  HashTable<SearchSymbol, Long>& hash_table =
    search_levels_a(last_level).getSymbolHashTable();
  
  Vector<StatisticalModel>& stat_models =
    search_levels_a(last_level).getStatisticalModels();
  
  // if no statistical model is found, output a warning message
  //
  if(stat_models.length() == 0) {
    Console::put(L"Warning: no statistical models available in the last search level.");
  }
  
  // if statistical models are available
  // store the statistical models into a separate SOF file
  //
  else {
    
    // open the SOF file
    //
    Sof stat_sof;

    // open the model file for i/o (TEXT)
    //
    if (type_a == TEXT) {
      
      if(!stat_sof.open(stat_file_a, File::WRITE_ONLY)) {
	return Error::handle(stat_file_a, L"open", Error::TEST, __FILE__, __LINE__);
      }
    }

    // open the model file for i/o (BINATY)
    //
    if (type_a == BINARY) {
      
      if(!stat_sof.open(stat_file_a, File::WRITE_ONLY, File::BINARY)) {
	return Error::handle(stat_file_a, L"open", Error::TEST, __FILE__, __LINE__);
      }
    }

    // store the hashtable and statistical models into the statistical SOF file
    //
    hash_table.write(stat_sof, last_level, PARAM_STAT_HASH);
    stat_models.write(stat_sof, last_level, PARAM_STAT);
    
    // close the statistical model SOF file
    //
    stat_sof.close();
  }
    
  // exit gracefully
  //
  return true;
}

// method: store
//
// arguments:
//  const String& model_file_a: (output) file name for a SOF file containing
//                          language models
//  const Vector<SearchLevel>& search_levels_a: (input) a set of search levels
//  OUTPUT_FORMAT format_a: (input) user-defined file format for the output model file
//  OUTPUT_TYPE type_a: (input) user-defined file type (text/binary)
//
//
// return: a boolean value indicating status
//
// this method outputs a set of input search levels to two SOF files
// with user-defined filenames according to a user-chosen file format
// that is either NATIVE format or JSGF format
//
boolean LanguageModel::store(const String& model_file_a,
			     Vector<SearchLevel>& search_levels_a,
			     OUTPUT_FORMAT format_a, OUTPUT_TYPE type_a) {
  // set an empty acoustic model file
  //
  String ac_model;
  return store (model_file_a, ac_model, search_levels_a, format_a, type_a);
}

// method: convertToJSGF
//
// arguments:
//   DiGraph<SearchNode> graph_a: (input) a DiGraph object to be converted
//   SearchSymbol grammar_name_a: (input) the JSGF grammar name for the graph
//
// return: (output) a JSGF grammar string converted from DiGraph
//
// This method converts the input grammar from DiGraph of SearchNode format
// to JSGF format. Each SearchSymbol is picked up from the input symbol table
// according to the SearchNode index in the graph.
//                 
String LanguageModel::convertToJSGF(DiGraph<SearchNode>& graph_a,
				    SearchSymbol grammar_name_a) {
  
  // declare and initialize a JSGF grammar string
  //
  String JSGF_output(L"  // Define the grammar name\n");
  JSGF_output.concat(L"  grammar network.grammar.");
  JSGF_output.concat(grammar_name_a);
  JSGF_output.concat(L";\n\n");
  JSGF_output.concat(L"  // Define the rules\n");
  JSGF_output.concat(L"  public");

  String debug_string;

  // debug message
  //
  debug_string.assign(L"debug 1");
  debug_string.debug(L"point");
  
  // get all vertices in the input graph
  //
  long num_vertices = graph_a.length() + 2;
  GraphVertex<SearchNode>* verts[num_vertices];
  
  verts[0] = graph_a.getStart();
  
  graph_a.gotoFirst();
  for(long i = 1; i < num_vertices - 1; i++) {
    verts[i] = (GraphVertex<SearchNode>*)graph_a.getCurr();
    graph_a.gotoNext();
  }
  
  verts[num_vertices - 1] = graph_a.getTerm();
  String rule; 
  SingleLinkedList<String> open_list;
  String start(L"ISIP_JSGF_1_0_START");
  open_list.insert(&start);

  Vector<Triple<String, SingleLinkedList<Long>, Float> > close_list;

  SingleLinkedList<Long> rule_vertex_list;
  Long start_index(0);
  rule_vertex_list.insert(&start_index);
  Triple<String, SingleLinkedList<Long>, Float > triple;
  Float rule_weight;
  rule_weight.assign(100000); 
  triple.assign(start, rule_vertex_list, rule_weight);
  close_list.concat(triple);

  long exten_index = 0;
  
  // loop through all rules
  //
  boolean more_rules = open_list.gotoFirst();
  while(more_rules) {

    // debug message
    //
    debug_string.assign(L"debug 2");
    debug_string.debug(L"point");
  
    // getting current rule name;
    //
    String temp_string;
    String* rulename = &temp_string;
    //String* rulename = NULL;
    open_list.remove(rulename);
    
    if(rulename->eq(L"ISIP_JSGF_1_0_START")) {
      rule.concat(L" <");
      rule.concat(grammar_name_a);  
      rule.concat(L"> = ");
    }
    else if(rulename->eq(L"ISIP_JSGF_1_0_TERM")) {
      more_rules = open_list.gotoFirst(); 

      // debug message
      //
      debug_string.assign(L"debug 3");
      debug_string.debug(L"point");
  
      continue;
    }
    else {
      rule.concat(L" <");
      rule.concat(*rulename);  
      rule.concat(L"> = ");
    }

    long i_index = 0;
    for(long i = 0; i < close_list.length(); i++) {
      if(rulename->eq(close_list(i).first())) {
	 rule_vertex_list = close_list(i).second();
	 i_index = i;
	 break;
      }
    }
    rule_weight = close_list(i_index).third();

    int more_vertex = 0;
    SearchSymbol tmp_sym;
    SearchSymbol sym;
    //GraphVertex<SearchNode>* curr_vert = new GraphVertex<SearchNode>();
    GraphVertex<SearchNode>* curr_vert = (GraphVertex<SearchNode>*) NULL;
    
    // debug message
    //
    debug_string.assign(L"debug 5");
    debug_string.debug(L"point");
  
    // loop through all vertices created by JSGF rule
    //
    for(boolean more_vertices = rule_vertex_list.gotoFirst();
	more_vertices; more_vertices = rule_vertex_list.gotoNext()) {
    
      // getting current vertex 
      //
      Long * curr_vert_index = rule_vertex_list.getCurr();
      curr_vert = verts[(long)*curr_vert_index];
     
      if (curr_vert->getItem() == &DiGraph<SearchNode>::START_OBJ) {
	sym.assign(L"<ISIP_JSGF_1_0_START>");
      }
      else if (curr_vert->getItem() == &DiGraph<SearchNode>::TERM_OBJ) {

	// debug message
	//
	debug_string.assign(L"debug 6");
	debug_string.debug(L"point");
  
	continue;
      }
      else {
	if(more_vertex==0) {
	  curr_vert->getItem()->getSymbol(sym);
	  tmp_sym.concat(sym);
	  more_vertex++;
	}
	else {
	  tmp_sym.concat(L" | ");
	  curr_vert->getItem()->getSymbol(sym);
	  tmp_sym.concat(sym);
	  more_vertex++;
	}
      }
      
      // debug message
      //
      debug_string.assign(L"debug 7");
      debug_string.debug(L"point");
        
    }// end for(more_vertices)

    
    if (more_vertex <= 1)     
      rule.concat(sym);
    else {
      rule.concat(L"( ");
      rule.concat(tmp_sym);
      rule.concat(L" )");
    }

    rule.concat(L" (");

    int is_alternative = 0;
    
    if((float)rule_weight <= 0) {

      // the rule is self loop
      //
      String tmp_weight;
      tmp_weight.assign(rule_weight);
      
      // add weight to the rule line
      //
      rule.concat(L" /");
      rule.concat(tmp_weight);
      rule.concat(L"/ ");
      rule.concat(L"<");
      rule.concat(*rulename);
      rule.concat(L">");
      is_alternative++;
    }

    SingleLinkedList<GraphArc<SearchNode> >subgroup_list;

    Vector<String> other_rules;
    
    // loop through all the available arcs starting from the current vertex
    //
    for (boolean more_paths = curr_vert->gotoFirst(); more_paths;
	 more_paths = curr_vert->gotoNext()) {

      // debug message
      //
      debug_string.assign(L"debug 8");
      debug_string.debug(L"point");
        
      // get the next vertex and its weight
      //
      GraphArc<SearchNode>* tmp_arc = curr_vert->getCurr();
      GraphVertex<SearchNode>* next_vert = tmp_arc->getVertex();

      boolean selfrule = false;
      boolean otherrule = false;
      String tmp_name;
      
      for(long i = 0; i < close_list.length(); i++) {
	SingleLinkedList<Long> vertex_list = close_list(i).second();
	boolean is_equal_vertex = false;

      // debug message
      //
      debug_string.assign(L"debug 9");
      debug_string.debug(L"point");
        
	for(boolean more_vertices = vertex_list.gotoFirst();
	    more_vertices; more_vertices = vertex_list.gotoNext()) {
      // debug message
      //
      debug_string.assign(L"debug 10");
      debug_string.debug(L"point");
        
	  Long * tmp_vert_index = vertex_list.getCurr();
	  GraphVertex<SearchNode>* tmp_vert = verts[(long)*tmp_vert_index];
	  if (next_vert->eq(*tmp_vert) && next_vert->compareVertices(*tmp_vert)) {
	    is_equal_vertex = true;
	    break;
	  }
	}
	if (is_equal_vertex) {
	  tmp_name = close_list(i).first();
	  if (tmp_name.eq(*rulename)) {	     
	    selfrule = true;
	    break;
	  }
	  else {
	    otherrule = true;
	    break;
	  }
	}
      }//end for(long i =0)
      
      if (selfrule) {
	
	// do nothing
	//	
      }
      else if (otherrule) {
	boolean done_other_rules = false;
	for( long r_index = 0; r_index < other_rules.length(); r_index++) {
	  if(tmp_name.eq(other_rules(r_index))) {
	    done_other_rules = true;
	  }
	}
	if(done_other_rules) {

	  //do nothing
	  //
	}
	else {  
	  float weight = tmp_arc->getWeight();

	  if(is_alternative == 0) {
	    is_alternative++;
	  }
	  else {
	    rule.concat(L" | ");
	  }
	  
	  // add weight to the rule line
	  //
	  rule.concat(L" /");
	  rule.concat(weight);
	  rule.concat(L"/ ");
	  rule.concat(L" <");
	  rule.concat(tmp_name);
	  rule.concat(L"> ");
	  other_rules.concat(tmp_name);
	}
      }
      else {

	// added to singlinkedlist to get new subgroup
	//
	subgroup_list.insert(tmp_arc);
      }	  
    }// for(more paths)

    // getting subgroup
    //
    Vector<SingleLinkedList<GraphArc<SearchNode> > >subgroup_table;

    getSubgroup(subgroup_list, subgroup_table, verts, num_vertices);
    
    // write subgroup information to open_list and close_list
    //
    for(long k = 0; k < subgroup_table.length(); k++) {
      
      SingleLinkedList<GraphArc<SearchNode> > new_group;
      new_group = subgroup_table(k);
      new_group.gotoFirst();
            
      // getting current rule name;
      //
      GraphArc<SearchNode>* current_arc;
      current_arc = new_group.getCurr();
      GraphVertex<SearchNode>* current_vert = current_arc->getVertex();
      float weight = current_arc->getWeight();
      
      boolean is_loop_extension = false;
      Float self_weight = 100000;
      
      // debug message
      //
      debug_string.assign(L"debug 11");
      debug_string.debug(L"point");
        
      // loop through all the available arcs starting from the current vertex
      //
      for (boolean more_paths = current_vert->gotoFirst(); more_paths;
	   more_paths = curr_vert->gotoNext()) {
	
      // debug message
      //
      debug_string.assign(L"debug 12");
      debug_string.debug(L"point");
        
	// get the next vertex and its weight
	//
	GraphArc<SearchNode>* tmp_arc = current_vert->getCurr();
	GraphVertex<SearchNode>* next_vert = tmp_arc->getVertex();
	if(current_vert->eq(*next_vert)) {
	  is_loop_extension = true;
	  self_weight = tmp_arc->getWeight();
	  break;
	}
      }
      
      // get rule extension
      //
      if(is_alternative == 0) {
	is_alternative++;
      }
      else {
	rule.concat(L" | ");
      }
      
      String exten;
      String extension_name;
      
      // add extension to the rule line
      //
      if(current_vert->getItem() == &DiGraph<SearchNode>::TERM_OBJ) {
	exten.concat(L" <ISIP_JSGF_1_0_TERM>");
	extension_name.concat(L"ISIP_JSGF_1_0_TERM");
      }
      else {		
	exten.concat(L" <extension");
	exten.concat(exten_index);
	exten.concat(L">");
	extension_name.concat(L"extension");
	extension_name.concat(exten_index);
	exten_index++;
      }
            
      // add weight to the rule line
      //
      rule.concat(L" /");
      rule.concat(weight);
      rule.concat(L"/ ");	
      rule.concat(exten);
      open_list.insert(&extension_name);
      SingleLinkedList<Long> vertex_list;

      // loop through all the available arcs starting from the current vertex
      //
      for (boolean more_paths = new_group.gotoFirst(); more_paths;
	   more_paths = new_group.gotoNext()) {

      // debug message
      //
      debug_string.assign(L"debug 13");
      debug_string.debug(L"point");
        
	// get the next vertex and its weight
	//
	GraphArc<SearchNode>* tmp_arc = new_group.getCurr();
	GraphVertex<SearchNode>* next_vert = tmp_arc->getVertex();
	Long index;
	for (long j = 0; j < num_vertices; j++) {
	  if ( next_vert == verts[j]) {
	    index = j;
	    break;
	  }
	}
	vertex_list.insert(&index);
      }
	
      // store runle_extension information into close_list and open_list
      //
      Triple<String, SingleLinkedList<Long>, Float > triple;          
      triple.assign(extension_name, vertex_list, self_weight);
      close_list.concat(triple);  	
    }
       
    rule.concat(L" ) "); 
    rule.concat(L";\n");
    more_rules = open_list.gotoFirst();     
  }// end for(more_rules)
  
  JSGF_output.concat(rule);

  // return the JSGF grammar string
  //
  return JSGF_output;
}

// method: getSubgroup
//
// arguments:SingleLinkedList<GraphArc<SearchNode> > & subgroup_list_a: (input)//  a list of GraphArc 
//  Queue<SingleLinkedList<GraphArc<SearchNode> > > &subgroup_table_a: (output)
//  a Vector containing subgroup based on weights and connection relationship
//  GraphVertex<SearchNode>** verts_a; (input) an array of GraphVertex storing
//  information of graph
//  long num_vertices_a; (input) number of vertex of the graph
// return: a boolean value indicating status
//
// This method takes a GraphArc list as input and based on weights and
// connection relationship in the graph, group GraoupArc to form new subgroup
//   
boolean LanguageModel::getSubgroup(SingleLinkedList<GraphArc<SearchNode> >
				   subgroup_list_a, Vector<SingleLinkedList<GraphArc<SearchNode> > >& subgroup_table_a, GraphVertex<SearchNode>* verts_a[],
				   long num_vertices_a){

  // debug message
  //
  String debug_string;
  
  // loop through all GraphArcs
  //
  for(boolean more_arcs = subgroup_list_a.gotoFirst();
      more_arcs; more_arcs = subgroup_list_a.gotoNext()) {

      // debug message
      //
      debug_string.assign(L"debug 14");
      debug_string.debug(L"point");
        
    // getting current rule name;
    //
    GraphArc<SearchNode>* current_arc;
    current_arc = subgroup_list_a.getCurr();
    
    // check if the arc has been processed and put into subgroup_table_a
    //
    boolean is_subgroup = false;
    for(long i = 0; i < subgroup_table_a.length(); i++) {
       SingleLinkedList<GraphArc<SearchNode> > tmp_group;
       tmp_group = subgroup_table_a(i);

      // debug message
      //
      debug_string.assign(L"debug 15");
      debug_string.debug(L"point");
        
       // loop through all GraphArcs
       //
       for(boolean more_arcs = tmp_group.gotoFirst();
	   more_arcs; more_arcs = tmp_group.gotoNext()) {

      // debug message
      //
      debug_string.assign(L"debug 16");
      debug_string.debug(L"point");
        
	 // get current arc
	 //
	 GraphArc<SearchNode>* tmp_arc;
	 tmp_arc = tmp_group.getCurr();
	 if(current_arc->eq(*tmp_arc)) {
	   is_subgroup = true;
	   break;
	 }    
       }
       if(is_subgroup)
	 break;       
    }
    
    if(is_subgroup) {      

      // debug message
      //
      debug_string.assign(L"debug 17");
      debug_string.debug(L"point");
        
      continue;
    }
    
   SingleLinkedList<GraphArc<SearchNode> > new_group;
    new_group.insert(current_arc);
    float weight = current_arc->getWeight();
    GraphVertex<SearchNode>* current_vert = current_arc->getVertex();
    boolean more_a = subgroup_list_a.gotoNext();
    while(more_a) {
      GraphArc<SearchNode>* next_arc;
      next_arc = subgroup_list_a.getCurr();
      GraphVertex<SearchNode>* next_vert = next_arc->getVertex();
      float next_weight = next_arc->getWeight();
      if(weight == next_weight && current_vert->compareVertices(*next_vert)) {
	new_group.insert(next_arc);
	//subgroup_list_a.remove(next_arc);
      }
      more_a = subgroup_list_a.gotoNext();
    }// end while(more_a)
    subgroup_table_a.concat(new_group);
    more_a = subgroup_list_a.gotoFirst();
  }// end for(more_arcs)
   
  Vector<SingleLinkedList<GraphArc<SearchNode> > > subgroup_table_tmp;

  Long num_subgroup = subgroup_table_a.length();
    
  // loop through all new group
  //
  for (long k =0; k < subgroup_table_a.length(); k++) {
    SingleLinkedList<GraphArc<SearchNode> > new_group;
    new_group = subgroup_table_a(k);
    
      // debug message
      //
      debug_string.assign(L"debug 18");
      debug_string.debug(L"point");
        
    // loop through all GraphArcs
    //
    for(boolean more_arcs = new_group.gotoFirst();
	more_arcs; more_arcs = new_group.gotoNext()) {

      // debug message
      //
      debug_string.assign(L"debug 19");
      debug_string.debug(L"point");
        
      // getting current rule name;
      //
      GraphArc<SearchNode>* current_arc;
      current_arc = new_group.getCurr();
      
      SingleLinkedList<GraphArc<SearchNode> > new_sub_group;
      new_sub_group.insert(current_arc);

      GraphVertex<SearchNode>* current_vert = current_arc->getVertex();
      SingleLinkedList<Long> common_parent;
      Vector<Float> common_weight;
      for(long i = 0; i < num_vertices_a; i++){
	
      // debug message
      //
      debug_string.assign(L"debug 20");
      debug_string.debug(L"point");
        
	if(!(verts_a[i])->eq(*current_vert)) {
	  // loop through all the available arcs starting from the current vertex
	  //
	  for (boolean more_paths = verts_a[i]->gotoFirst(); more_paths;
	       more_paths = verts_a[i]->gotoNext()) {

      // debug message
      //
      debug_string.assign(L"debug 21");
      debug_string.debug(L"point");
        
	    // get the next vertex and its weight
	    //
	    GraphArc<SearchNode>* tmp_arc = verts_a[i]->getCurr();
	    GraphVertex<SearchNode>* next_vert = tmp_arc->getVertex();
	    float weight = tmp_arc->getWeight();
	    if(next_vert->eq(*current_vert)) {
	      Long index(i);
	      common_parent.insert(&index);
	      Float tmp_weight;
	      tmp_weight.assign(weight);
	      common_weight.concat(tmp_weight);
	      break;
	    }
	  }// end for( boolean more_paths)
	}// end if(!(verts_a[i])
      }
      
      boolean more_ab = new_group.gotoNext();  
      while(more_ab) {
      // debug message
      //
      debug_string.assign(L"debug 22");
      debug_string.debug(L"point");
        
	GraphArc<SearchNode>* next_arc;
	next_arc = new_group.getCurr();
	GraphVertex<SearchNode>* next_vert = next_arc->getVertex();
	long j, match;
	j =0;
	match = 0;
	
	// checking if current_vert and next_vert has common parent and weight
	//
	for(boolean more_p = common_parent.gotoFirst();
	    more_p; more_p = common_parent.gotoNext()) {
	  GraphVertex<SearchNode>* common_vertex;
	  Long *common_vertex_index;
	  common_vertex_index = common_parent.getCurr();
	  common_vertex = verts_a[(long)*common_vertex_index];
	  for (boolean more_paths = common_vertex->gotoFirst(); more_paths;
	       more_paths = common_vertex->gotoNext()) {
	    
	    // get the next vertex and its weight
	    //
	    GraphArc<SearchNode>* tmp_arc = common_vertex->getCurr();
	    GraphVertex<SearchNode>* tmp_vert = tmp_arc->getVertex();
	    float weight = tmp_arc->getWeight();
	    if(next_vert->eq(*tmp_vert) && common_weight(j).eq(weight)) {
	      match++;
	      break;
	    }
	  }// end for( boolean more_paths)
	  j++;
	}
	
	if(j == match ) {
      // debug message
      //
      debug_string.assign(L"debug 23");
      debug_string.debug(L"point");
        
	  new_sub_group.insert(next_arc);
	  //new_group.remove(next_arc);
	}
      // debug message
      //
      debug_string.assign(L"debug 24");
      debug_string.debug(L"point");
        
	more_ab = new_group.gotoNext();  
      }// end while(more_ab)
      // debug message
      //
      debug_string.assign(L"debug 25");
      debug_string.debug(L"point");
        
      subgroup_table_tmp.concat(new_sub_group);
    } //end for)boolean more_arcs)
  }// end for(!subgroup_table_a)

 
  // exit gracefully
  //
  return true;
}


// method: digraphToJSGF
//
// arguments:
//   DiGraph<SearchNode> graph_a: (input) a DiGraph object to be converted
//   SearchSymbol grammar_name_a: (input) the JSGF grammar name for the graph
//
// return: (output) a JSGF grammar string converted from DiGraph
//
// This method converts the input grammar from NATIVE format
// to JSGF format. Each SearchSymbol is picked up from the input symbol table
// according to the SearchNode index in the graph.
//                 
String LanguageModel::digraphToJSGF(DiGraph<SearchNode>& graph_a,
				    SearchSymbol grammar_name_a) {
  
  // declare and initialize a JSGF grammar string
  //
  String JSGF_output(L"  // Define the grammar name\n");
  JSGF_output.concat(L"  grammar network.grammar.");
  JSGF_output.concat(grammar_name_a);
  JSGF_output.concat(L";\n\n");
  JSGF_output.concat(L"  // Define the rules\n");
  JSGF_output.concat(L"  public");

  String debug_string;

  // get all vertices in the input graph
  //
  long num_vertices = graph_a.length() + 2;
  GraphVertex<SearchNode>* verts[num_vertices];

  // set the first vertex
  //
  verts[0] = graph_a.getStart();
  
  graph_a.gotoFirst();
  for(long i = 1; i < num_vertices - 1; i++) {
    verts[i] = (GraphVertex<SearchNode>*)graph_a.getCurr();
    graph_a.gotoNext();
  }

  // set the last vertex
  //
  verts[num_vertices - 1] = graph_a.getTerm();

  // set the rules names
  //
  String rule_names[num_vertices];

  // init the initial rule names
  //
  rule_names[0].assign(L"ISIP_JSGF_1_0_START");
  rule_names[num_vertices - 1].assign(L"ISIP_JSGF_1_0_TERM");

  for (int i = 1; i < num_vertices -1 ; i++){
    rule_names[i].assign(L"extention");
    rule_names[i].concat((long)i);
  }

  // set the rule bodies
  //
  String rule_bodies[num_vertices];
  GraphVertex<SearchNode>* curr_vert = (GraphVertex<SearchNode>*) NULL;
  
  for (int i = 0; i < num_vertices; i++){

    // get current vertex
    //
    curr_vert = verts[i];

    // set the symbol name
    //
    SearchSymbol symbol_name;
    if (verts[i]->isStart()){
      symbol_name.assign(L"<ISIP_JSGF_1_0_START>");
    }
    else if (verts[i]->isTerm()){
      symbol_name.assign(L"<ISIP_JSGF_1_0_TERM>");
    }
    else{
      verts[i]->getItem()->getSymbol(symbol_name);
    }
    
    rule_bodies[i].assign(symbol_name);
    
    // seperate the symbol name from
    // following tokens
    //
    rule_bodies[i].concat(L" ");

    String vertices_rule;
    
    int flag = 0;
    
    // loop through all the available arcs starting from the current vertex
    //
    for (boolean more_paths = curr_vert->gotoFirst(); more_paths;
	 more_paths = curr_vert->gotoNext()) {

      // get current arc
      //
      GraphArc<SearchNode>* curr_arc = curr_vert->getCurr();
      float weight = curr_arc->getWeight();
      GraphVertex<SearchNode>* tmp_vert = curr_arc->getVertex();

      int index = -1;
      for (index = 0; index < num_vertices; index++){
	if ( verts[index] == tmp_vert ){
	  break;
	}
      }

      String rule_name = rule_names[index];
      
      // write the rule
      //
      // if it is the first one
      //
      if ( flag == 0){

	vertices_rule.concat(L"( /");
      }
      else {
	vertices_rule.concat(L"| ( /");	
      }
      
      vertices_rule.concat(weight);
      vertices_rule.concat(L"/ <");
      vertices_rule.concat(rule_name);	
      vertices_rule.concat(L"> )");
      flag++;
    }

    // group all next vertices
    //
    if ( flag > 1 ){
      String tmp_body(L"(");
      tmp_body.concat(vertices_rule);
      tmp_body.concat(L")");
      rule_bodies[i].concat(tmp_body);
    }
    else {
      rule_bodies[i].concat(vertices_rule);
    }
    
  }

  
  
  // write the whole rule
  //
  for (int i = 0; i < num_vertices; i++){

    // set the rule name
    //
    String rule_name;
    
    if (verts[i]->isStart()){
      rule_name.assign(grammar_name_a);
    }
    else if (verts[i]->isTerm()){
      continue;
    }
    else{
      rule_name = rule_names[i];
    }
    
    JSGF_output.concat(L"  ");
    JSGF_output.concat(L"<");
    JSGF_output.concat(rule_name);
    JSGF_output.concat(L"> = ");
    JSGF_output.concat(rule_bodies[i]);
    JSGF_output.concat(L" ;\n");
    
  }
  
  // return the JSGF grammar string
  //
  return JSGF_output;
}
