// file: $isip/class/search/StackSearch/ssrch_11.cc
// version: $Id: ssrch_11.cc,v 1.3 2002/08/02 02:07:17 jelinek Exp $
//

// isip include files
//
#include "StackSearch.h"

// method: traverseTraceLevels
//
// arguments: none
//
// return: logical error status
//
// do a single-step traversal of the hypotheses at each level
//
boolean StackSearch::traverseTraceLevels() {
  
  // define local variables
  //
  boolean status = false;
  
  // make sure we have some paths left to traverse
  //
  if (!pathsRemainTrace()) {
    return false;
  }

  // propagate traces forward until they are all ready to be evaluated
  //
  status = (propagateTraces() || status);

  // now we evaluate all traces that are in the lowest level and move 
  // them forward one arc. Viterbi pruning takes place at this level 
  // as does beam pruning.
  //
  status = (evaluateTraceModels() || status);

  // exit gracefully
  //
  return status;
}

// method: propagateTraces
//
// arguments:
//  DoubleLinkedList<Trace>* top_list: (output) traces that reached top level
//
// return: logical error status
//
// move traces through the hierarchy until all active traces are ready to be
// evaluated
//
boolean StackSearch::propagateTraces(DoubleLinkedList<Trace>* top_list_a) {
  
  // define local variables
  //
  long level_num = 0;
  boolean status = true;
  long num_levels = getNumLevels();
  String out;
  
  // clear the hypothesis list since we are generating new hypotheses for this
  // frame
  //
  clearValidHypsTrace();

  // move all traces forward in the search space. we do this by
  // pushing traces up and down in the hierarchy until they all rest
  // in a state that is ready to be evaluated. if there is an infinite
  // "skip" loop in the search graph then this process will go into an
  // infinite loop. an infinite skip loop is of the form:
  // A -> B && B -> A.
  // this infinite loop structure may be masked by the fact that it 
  // extends over multiple search levels.
  // 
  // when we exit the while loop below, all traces should be in the lowest
  // level and should be ready for evaluation.
  //
  while ((getActiveTraces() != getActiveTraces(num_levels - 1)) ||
	 status) {
    
    // status indicates whether any movement in the traces happened
    //
    status = false;
    
    // work from the bottom up until we reach a point where all traces
    // are ready to descend again. this segment takes care of
    // posterior scores such as nsymbol probabilities and applies
    // pruning (beam and instance) at each level.
    //
    for (level_num = num_levels - 1; level_num >= 0; level_num--) {
      
      // propagate up
      //
      status = (propagateTracesUp(level_num, top_list_a) || status);
    }
    
    for (level_num = 0;  level_num < (num_levels - 1); level_num++) {
      if (getSearchLevel(level_num).useBeam()) {
	if (debug_level_d >= Integral::DETAILED) {
	  Console::put(L"\npruning after propagation traces up");
	}
	beamPruneTrace(level_num);
      }
    }
    
    // propagate traces down the hierarchy, carrying out Viterbi pruning
    // as we go.
    //
    for (level_num = 0;  level_num < num_levels - 1; level_num++) {
      
      // propagate down
      //
      status = (propagateTracesDown(level_num) || status);
    }
  }
  
  // exit gracefully
  //
  return true;
}


// method: evaluateTraceModels
//
// arguments: (none)
//
// return: logical error status.
//
// evaluate all statistical models
//
boolean StackSearch::evaluateTraceModels() {

  // define local variables
  //
  String out;
  String val;
  long level_num = getNumLevels() - 1;
  boolean end_loop = false;
  Trace* curr_trace = (Trace*)NULL;
  Trace* next_trace = (Trace*)NULL;    
  History* curr_history = (History*)NULL;
  GraphVertex<SearchNode>* curr_vert = (GraphVertex<SearchNode>*)NULL;
  SearchSymbol item;
 
  // output the frame information
  //
  if (debug_level_d >= Integral::DETAILED) {
    out.assign(L"\nevaluating data in frame "); out.concat(current_frame_d);
    Console::put(out);
  }

  // loop over all active traces and score them
  //
  for (boolean more_traces = trace_lists_d(level_num).gotoFirst();
       more_traces; more_traces = trace_lists_d(level_num).gotoNext()) {
    
    // get the current vertex in the search graph
    //
    curr_trace = trace_lists_d(level_num).getCurr();
    curr_history = curr_trace->getHistory();
    curr_vert = curr_history->peek()->getCentralVertex();

    // evaluate the statistical model
    //
    SearchNode* p_sn = curr_vert->getItem();
    if (p_sn == NULL) {
      return Error::handle(name(),
			   L"evaluateTraceModels - no SearchNode in Vertex",
			   Error::ARG, __FILE__, __LINE__);
    }

    // find the statistical model index
    
    // note: this code is not effective because it needs to call a
    // hashing function to get the statistical model index, we plan to
    // use look up table access in the next version
    //
    // get the symbol id
    //
    long symbol_id = p_sn->getSymbolId();

    // get search level and symbol table
    //
    SearchLevel* sl = p_sn->getSearchLevel();
    Vector<SearchSymbol>& symbol_table = sl->getSymbolTable();
    SearchSymbol& ss = symbol_table(symbol_id);

    // get a hashtable hashing search symbols to statistical model indices
    //
    HashTable<SearchSymbol, Long>& symbol_hashtable =
      sl->getSymbolHashTable();

    // check if hash table is loaded
    //
    if (symbol_hashtable.getNumItems() == 0) {
      return Error::handle(name(),
			   L"evaluateTraceModels - load stat models hashtable first",
			   Error::ARG, __FILE__, __LINE__);
    }

    // get the statistical model index from hashtable
    //
    Long* p_stat_model_id = symbol_hashtable.get(ss);
    long stat_model_id = (long)*p_stat_model_id;

    // having the statistical model index, we can check whether the stat

    float stat_score = stat_model_scores_d(current_frame_d, stat_model_id);
    if (stat_score == SearchNode::INACTIVE_SCORE) {
      stat_score =
	p_sn->evaluateData(features_d(current_frame_d), current_frame_d);
      stat_model_scores_d.setValue(current_frame_d, stat_model_id, stat_score);
    }
    
    float new_score = curr_trace->getScore() + stat_score;
    curr_trace->setScore(new_score);
    
    // update the maximal trace score for the beam pruning
    //
    if (getSearchLevel(level_num).useBeam()
	&& (new_score > max_frame_scores_d(current_frame_d, level_num))) {

      if (debug_level_d >= Integral::DETAILED) {
		
	// print the score update information
	//
	out.assign(L"fr:");
	out.concat(current_frame_d);
	out.concat(L" lv:");
	out.concat(level_num);
	out.concat(L" max_sc updt (after the model EVALUATION) ");
	out.concat(max_frame_scores_d(current_frame_d, level_num));
	out.concat(L"->");
	out.concat(new_score);
	Console::put(out);
      }

      // update the maximal trace score
      //
      max_frame_scores_d.setValue(current_frame_d, level_num, new_score);
    }
  }

  // do the beam pruning at this level
  //
  if (getSearchLevel(getNumLevels() - 1).useBeam()) {
    if (debug_level_d >= Integral::DETAILED) {
      Console::put(L"\npruning after the model EVALUATION");
    }
    beamPruneTrace(getNumLevels() - 1);
  }
  
  // update current frame
  //        - traces generated later will correspond to the next frame
  //
  current_frame_d += 1 ;

  if (debug_level_d >= Integral::DETAILED) {
    out.assign(L"\nTIME CHANGED to: ");
    out.concat(current_frame_d);
    Console::put(out);
  }
  
  // we mark the current end of the list so we only propagate those traces
  // forward that were just evaluated
  //
  trace_lists_d(level_num).gotoLast();
  trace_lists_d(level_num).setMark();
  trace_lists_d(level_num).gotoFirst();

  if (debug_level_d >= Integral::DETAILED) {
    Console::put(L"\npropagating traces forward after the model EVALUATION");
  }
  
  // loop until we have propagated all evaluated traces forward
  //
  end_loop = (trace_lists_d(level_num).length() < 1);
  while (!end_loop) {
    
    end_loop = trace_lists_d(level_num).isMarkedElement();
    
    // get the current vertex in the search graph
    //
    curr_trace = trace_lists_d(level_num).getCurr();
    curr_history = curr_trace->getHistory();
    curr_vert = curr_history->peek()->getLastVertex();
    
    // if this trace is inactive, then delete it
    //
    if (!curr_trace->isActive()) {
      
      // bump the current trace off of the trace list
      //
      if (debug_level_d >= Integral::ALL) {
	Console::put(L" not active");
      }      
      trace_lists_d(level_num).removeFirst(curr_trace);
      Trace::deleteTrace(curr_trace, true);
    }
    
    // else, extend paths and apply viterbi pruning
    //
    else {
      
      // extend all paths from that vertex and apply viterbi pruning
      //
      for (boolean more_paths = curr_vert->gotoFirst(); more_paths;
	   more_paths = curr_vert->gotoNext()) {
	
	// create a new trace
	//
	next_trace = new Trace(*curr_trace);
	next_trace->setFrame((ulong)current_frame_d);
	
	// update the score with the arc score
	//
	GraphArc<SearchNode>* tmp_arc = curr_vert->getCurr();
	next_trace->setScore(next_trace->getScore() + tmp_arc->getWeight());
	
	// update the history with the next node at this level
	//
	next_trace->getHistory()->peek()->
	  assignAndAdvance((ulong)tmp_arc->getVertex());	
	
	// add the trace to the search node's trace list
	//
	Trace* existing_trace = (Trace*)NULL;
	SearchNode* next_node = next_trace->getHistory()->peek()->
	  getCentralVertex()->getItem();
 
	if (next_node->addTrace(next_trace, current_frame_d, existing_trace)) {
	  
	  // set the backpointer
	  //
	  next_trace->setBackPointer(curr_trace);
	
	  // add the trace to this level's trace list
	  //
	  trace_lists_d(level_num).insertLast(next_trace);

	  if (debug_level_d >= Integral::ALL) { 
	    printNewPath(next_trace, curr_trace);
	  }
	}
	else {

	  // delete the trace since there is an existing trace - same history
	  //
	  if (debug_level_d >= Integral::ALL) { 
	    printDeletedPath(next_trace, curr_trace);
	  }	  
	  delete next_trace;
	}
      }

      // bump the current trace off of the trace list
      //
      if (curr_trace->getRefCount() < 1) {
	Trace::deleteTrace(curr_trace, true);
      }
      trace_lists_d(level_num).removeFirst(curr_trace);
      trace_lists_d(level_num).gotoFirst();
      
    } // end if the trace is active
  }   // end while (!end_loop)
  
  // exit gracefully
  //
  return true;
}


// method: propagateTracesUp
//
// arguments:
//  long level_num: (input) the level for which we will propagate traces
//  DoubleLinkedList<Trace>* top_list: (output) traces that reached top level
//
// return: logical error status. if no propagations were made then return false
//
// propagate all traces up. only traces at the end of a subgraph are considered
//
boolean StackSearch::propagateTracesUp(long level_num_a,
		       DoubleLinkedList<Trace>* top_list_a) {

  // define local variables
  //
  boolean status = false;
  boolean end_loop = false;
  Trace* curr_trace = (Trace*)NULL;
  Trace* tmp_trace = (Trace*)NULL;
  Trace* next_trace = (Trace*)NULL;    
  History* curr_history = (History*)NULL;
  GraphVertex<SearchNode>* curr_vert = (GraphVertex<SearchNode>*)NULL;
  GraphVertex<SearchNode>* tmp_vert = (GraphVertex<SearchNode>*)NULL;
  Context useless;

  // output the debugging information
  //
  if (debug_level_d >= Integral::ALL) {
    String output(L"propagating traces up from level: ");
    output.concat(level_num_a); output.concat(L" in frame: ");
    output.concat(current_frame_d);
    Console::put(output);
  }
  
  // make sure there is at least one trace to propagate
  //
  if (trace_lists_d(level_num_a).length() < 1) {
    return false;
  }
  
  // loop over all current traces and create the next level's traces. all new
  // traces will go at the end of the list so we set the mark to tell us when
  // to stop propagating
  //
  trace_lists_d(level_num_a).gotoLast();
  trace_lists_d(level_num_a).setMark();
  trace_lists_d(level_num_a).gotoFirst();
  
  while (!end_loop) {
    
    // loop until we reached the marked element
    //
    end_loop = trace_lists_d(level_num_a).isMarkedElement();
    
    // get the current first trace from the list
    //
    trace_lists_d(level_num_a).removeFirst(curr_trace);
    
    // if the trace is inactive then delete it and its backpath if necessary
    //
    if (!curr_trace->isActive()) {
      Trace::deleteTrace(curr_trace, true);
    }
    
    // else, extend paths and update scores
    //
    else {
      
      // get the current vertex in the search graph
      //
      curr_history = curr_trace->getHistory();

      if (curr_history->length() == 0) {
	String out(L"zero history length for trace: ");
	out.concat(curr_trace);
	Console::put(out);
      }

      curr_vert = curr_history->peek()->getCentralVertex();
      
      // if this vertex is a terminal vertex then ascend and create the traces
      // at the higher level
      //
      if (curr_vert->isTerm()) {
	
	// set the status to true since we are propagating a trace
	//
	status = true;

	// if we are at the top level then put the trace into the valid
	// hypothesis list
	//
	if (level_num_a == 0) {
	  if (top_list_a != NULL) {
	    top_list_a->insertLast(curr_trace);
	  }
	  else {
	    valid_hyps_d.insertLast(curr_trace);
	  }
	}
	
	// else move up one level and create the next set of traces
	//
	else {
	  
	  // create a new trace, 
	  //
	  next_trace = new Trace(*curr_trace);
	  next_trace->setFrame((ulong)current_frame_d);
	  next_trace->setBackPointer(curr_trace);
	
	  // decrement the history of the new trace
	  //
	  if (!next_trace->getHistory()->isEmpty()) {
	    next_trace->getHistory()->pop(&useless);
	  }

	  if (debug_level_d >= Integral::ALL) {
	    printNewPath(next_trace, curr_trace);
	  }
		
	  // propagate all paths we can move to from this higher level trace
	  //
	  tmp_vert = next_trace->getHistory()->peek()->getLastVertex();

	  // if the vertex to propagate from is terminal, generate one trace,
	  // history will be shifted
	  //
	  if (tmp_vert->isTerm()) {

	    // create a temporary trace
	    //
	    tmp_trace = new Trace(*next_trace);
	    tmp_trace->setFrame((ulong)current_frame_d);

	    // shift the history and set the backpointer
	    //
	    tmp_trace->getHistory()->peek()->assignAndAdvance((ulong)tmp_vert);
	    tmp_trace->setBackPointer(next_trace);

	    // add the trace to this level's trace list (or top list)
	    //
	    if ((level_num_a - 1 == 0) && (top_list_a != NULL)) {
	      top_list_a->insertLast(tmp_trace);
	    }
	    else {
	      trace_lists_d(level_num_a - 1).insertLast(tmp_trace);
	    }

	    // print the debugging information
	    //
	    if (debug_level_d >= Integral::ALL) {
	      printNewPath(tmp_trace, next_trace);
	    }
	  }

	  // vertex to propagate from is not terminal
	  //
	  else {

	    // generate traces to all next vertices
	    //	  
	    for (boolean more_paths = tmp_vert->gotoFirst(); more_paths;
		 more_paths = tmp_vert->gotoNext()) {
	      
	      // create a new trace, we store the extra trace so we
	      // can later view the scores as they are added. this can
	      // be changed later for efficiency
	      //
	      tmp_trace = new Trace(*next_trace);
	      tmp_trace->setFrame((ulong)current_frame_d);
	      
	      // update the score with the arc score
	      //
	      GraphArc<SearchNode>* tmp_arc = tmp_vert->getCurr();
	      float tmp_score = tmp_trace->getScore() + tmp_arc->getWeight();
	      tmp_trace->setScore(tmp_score);
	    
	      // change the history
	      //
	      tmp_trace->getHistory()->peek()->
		assignAndAdvance((ulong)tmp_arc->getVertex());
	      
	      // add the trace to the search node's trace list
	      //
	      Trace* existing_trace = (Trace*)NULL;	    
	      SearchNode* next_node = tmp_trace->getHistory()->peek()->
		getCentralVertex()->getItem();
	      
	      if (next_node->addTrace(tmp_trace, current_frame_d,
				      existing_trace)) {
		
		// update the maximal trace score for the beam pruning
		//
		if (getSearchLevel(level_num_a - 1).useBeam()
		    && (tmp_score > max_frame_scores_d(current_frame_d, 
						       level_num_a - 1)) ) {
		  
		  if (debug_level_d >= Integral::DETAILED) {
		    
		    // print the debugging information
		    //
		    String output(L"fr:");
		    output.concat(current_frame_d);
		    output.concat(L" lv:");
		    output.concat(level_num_a - 1);
		    output.concat(L" max_sc updt (going to higher level YES) ");
		    output.concat(max_frame_scores_d(current_frame_d, 
						     level_num_a - 1));
		    output.concat(L"->");
		    output.concat(tmp_score);
		    Console::put(output);
		  }
		  
		  // update the maximal trace score
		  //
		  max_frame_scores_d.setValue(current_frame_d, level_num_a - 1,
					      tmp_score);
		}
	      
		// set the backpointer
		//
		tmp_trace->setBackPointer(next_trace);
		
		// add the trace to the higher level's trace list (or top list)
		//
		if ((level_num_a - 1 == 0) && (top_list_a != NULL)) {
		  top_list_a->insertLast(tmp_trace);
		}
		else {
		  trace_lists_d(level_num_a - 1).insertLast(tmp_trace);
		}

		if (debug_level_d >= Integral::ALL) {
		  printNewPath(tmp_trace, next_trace);
		}
	      }
	      else {
	      
		// delete the trace since there is existing trace -
		// same history
		//
		if (debug_level_d >= Integral::ALL) { 
		  printDeletedPath(tmp_trace, next_trace);
		}	  	      
		delete tmp_trace;
	      }
	    }
	  }
	  
	  // delete next_trace if no traces propagated from it have survived
	  //
	  if (next_trace->getRefCount() < 1) {
	    Trace::deleteTrace(next_trace, false);
	  }
	}
	
	// bump the current trace off of the trace list
	//
	if (curr_trace->getRefCount() < 1) {
	  
	  // if this is at top level, we can not delete them since they
	  // are also referenced by the valid_hyps
	  //
	  if (level_num_a != 0) {
	    Trace::deleteTrace(curr_trace, true);
	  }
	}
	trace_lists_d(level_num_a).gotoFirst();
      }
      
      // if this trace is not at the end of a subgraph then keep it on this
      // level
      //
      else {
	
	float curr_score = curr_trace->getScore();

	// update the maximal trace score for the beam pruning
	//
	if (getSearchLevel(level_num_a).useBeam()
	    && (curr_score > max_frame_scores_d(current_frame_d, level_num_a))
	    ) {

	  if (debug_level_d >= Integral::DETAILED) {

	    // print the debugging information
	    //
	    String output(L"fr:");
	    output.concat(current_frame_d);
	    output.concat(L" lv:");
	    output.concat(level_num_a);
	    output.concat(L" max_sc updt (going to higher level NO ) ");
	    output.concat(max_frame_scores_d(current_frame_d, level_num_a));
	    output.concat(L"->");
	    output.concat(curr_score);
	    Console::put(output);
	  }

	  // update the maximal trace score
	  //
	  max_frame_scores_d.setValue(current_frame_d, level_num_a,
				      curr_score);
	}
      
	// keep the trace on this level
	//
	trace_lists_d(level_num_a).insertLast(curr_trace);
	trace_lists_d(level_num_a).gotoFirst();
      }
    }
  }
  
  // exit gracefully
  //
  return status;
}


// method: propagateTracesDown
//
// arguments:
//  long level_num: (input) the level for which we will propagate traces
//
// return: logical error status. if no propagations were made then return false
//
// propagate all traces down. any traces that are at the end of a subgraph are
// discarded.
//
boolean StackSearch::propagateTracesDown(long level_num_a) {

  // define local variables
  //
  Trace* curr_trace = (Trace*)NULL;
  Trace* next_trace = (Trace*)NULL;    
  GraphVertex<SearchNode>* curr_vert = (GraphVertex<SearchNode>*)NULL;
  GraphVertex<SearchNode>* tmp_vert = (GraphVertex<SearchNode>*)NULL;
  float score = 0.0;
  SearchSymbol item;
  boolean status = false;

  // output the debugging information
  //
  if (debug_level_d >= Integral::ALL) {
    String output;        
    output.assign(L"propagating traces down from level: ");
    output.concat(level_num_a);
    output.concat(L" in frame: ");    
    output.concat(current_frame_d);
    Console::put(output);
  }

  // make sure there is at least one trace to propagate
  //
  if (trace_lists_d(level_num_a).length() < 1) {
    return false;
  }

  // switch on whether we are at the lowest level or not
  //
  // if we are not at the lowest level, push traces down the hierarchy -
  // there should be no traces left at this level when we are done
  //
  trace_lists_d(level_num_a).gotoFirst();
  
  if (level_num_a != getNumLevels() - 1) {
    
    // loop over all current traces and create the next level's traces
    //
    while (!trace_lists_d(level_num_a).isEmpty()) {
      
      // get the current vertex in the search graph
      //
      trace_lists_d(level_num_a).removeFirst(curr_trace);

      // delete this trace if it is now inactive
      //
      if (!curr_trace->isActive()) {

	// bump the current trace off of the trace list and delete it
	// and its backpath if necessary
	//
	Trace::deleteTrace(curr_trace, true);
      }

      // if it is active then propagate it
      //
      else {

	// get the current history and its central vertex
	//
	Context curr_context(*curr_trace->getHistory()->peek());
	curr_vert = curr_context.getCentralVertex();
	
	// if this vertex has a subgraph then descend and create the traces
	// at the next level
	//
	if (!curr_vert->isStart() && !curr_vert->isTerm()) {
	  
	  // we are propagating paths so set status to true
	  //
	  status = true;

	  // get the subgraph's start vertex
	  //
	  SearchLevel& sl_curr = getSearchLevel(level_num_a);
	  SearchLevel& sl_lower = getSearchLevel(level_num_a + 1);

	  // convert context vertices to search symbol indices
	  // we need to do it before hashing in getSubGraphIndex
	  //
	  curr_context.convert();
	  Ulong* subgr_ind = sl_curr.getSubGraphIndex(curr_context);
	  
	  if (subgr_ind == (Ulong*)NULL) {

	    // throw an error message and exit
	    //
	    String out(L"propagateTracesDown - no subgraph for context: ");
	    curr_context.debug(L"undefined context");
	    curr_trace->getHistory()->peek()->print();
	    Console::put(out);
	    return Error::handle(name(), L"propagateTracesDown", Error::ARG,
				 __FILE__, __LINE__);	    
	  }

	  // get the subgraph start vertex
	  // 
	  tmp_vert = (sl_lower.getSubGraph((long)*subgr_ind)).getStart();
	  
	  // loop over all paths in the subgraph and create new traces in the
	  // next level
	  //
	  for (boolean more_paths = tmp_vert->gotoFirst(); more_paths;
	       more_paths = tmp_vert->gotoNext()) {

	    long central_pos =
	      getSearchLevel(level_num_a + 1).getRightContext() + 1;
	    long left_context_length =
	      getSearchLevel(level_num_a + 1).getLeftContext();
	    long total_context_length = 
	      left_context_length + central_pos;
	    
	    // initialize left vertices of the context
	    //
	    Context left_context(total_context_length, central_pos);
	    for (long i = 0; i < left_context_length; i++) {
	      left_context.assignAndAdvance((ulong)tmp_vert);
	    }

	    // initialize central vertex of the context
	    //
	    GraphArc<SearchNode>* tmp_arc = tmp_vert->getCurr();
	    GraphVertex<SearchNode>* target_vertex = tmp_arc->getVertex();
	    left_context.assignAndAdvance((ulong)target_vertex);
	    
	    // output the debugging information
	    //
	    if (debug_level_d >= Integral::ALL) {
	      String out;
	      left_context.print();
	      Console::put(out);
	    }
	    
	    // generate list of all possible right contexts
	    //
	    DoubleLinkedList<Context> right_context_list;
	    right_context_list.setAllocationMode(DstrBase::USER);
	    long depth = getSearchLevel(level_num_a + 1).getRightContext();
	    generateRightContexts(right_context_list, left_context, depth);

	    // output the debugging information
	    //
	    if (debug_level_d >= Integral::ALL) {
	      String output(L"Generated contexts for depth");
	      output.concat(depth);
	      Console::put(output);
	      long num = 0;
	      for (boolean more_items = right_context_list.gotoFirst();
		  more_items;
		  more_items = right_context_list.gotoNext()) {
		num++;
		String out(L" ");
		out.concat(num);
		out.concat(L":");
		String cont;
		right_context_list.getCurr()->print();
		out.concat(cont);
		Console::put(out);
	      }
	    }

	    // generate a new trace for each right context
	    //
	    for (boolean more_items = right_context_list.gotoFirst();
		 more_items;
		 more_items = right_context_list.gotoNext()) {
	      
	      // create a new trace
	      //
	      next_trace = new Trace(*curr_trace);
	      next_trace->setFrame((ulong)current_frame_d);
	      
	      // update the score with the arc score
	      //
	      next_trace->setScore(next_trace->getScore() +
				   tmp_arc->getWeight());
	    
	      // add to the history
	      //
	      next_trace->getHistory()->push(right_context_list.getCurr());
	  	    
	      // add the trace to the search node's trace list.
	      //
	      Trace* existing_trace = (Trace*)NULL;
	      SearchNode* next_node = next_trace->getHistory()->peek()->getCentralVertex()->getItem();
	      
	      if (next_node->addTrace(next_trace, current_frame_d,
				      existing_trace)) {
		
		// set the backpointer
		//
		next_trace->setBackPointer(curr_trace);
		
		// add the trace to the next level's trace list
		//
		trace_lists_d(level_num_a + 1).insertLast(next_trace);

		if (debug_level_d >= Integral::ALL) {
		  printNewPath(next_trace, curr_trace);
		}
	      }
	      else {
		
		// delete the trace since there is existing trace - same history
		//
		if (debug_level_d >= Integral::ALL) { 
		  printDeletedPath(next_trace, curr_trace);
		}	  	      
		delete next_trace;
	      }
	    }

	    // clean the right contexts list
	    //
	    right_context_list.setAllocationMode(DstrBase::SYSTEM);
	    right_context_list.clear();
	  }
	    
	  // bump the current trace off of the trace list
	  //
	  if (curr_trace->getRefCount() < 1) {
	    Trace::deleteTrace(curr_trace, true);
	  }
	  trace_lists_d(level_num_a).gotoFirst();
	}
	
	// else if there is no subgraph for this node, we extend the search
	// along this level. for now, we don't allow a level to completely
	// skip sub-levels so if we get to the end of the graph at this level
	// we simply stop
	//
	else {
	  boolean more_paths = curr_vert->gotoFirst();
	  if (more_paths) {
	    status = true;
	  }
	  for ( ; more_paths; more_paths = curr_vert->gotoNext()) {
	    
	    // create a new trace
	    //
	    next_trace = new Trace(*curr_trace);
	    next_trace->setFrame((ulong)current_frame_d);
	    
	    // update the score with the arc score
	    //
	    GraphArc<SearchNode>* tmp_arc = curr_vert->getCurr();
	    score = next_trace->getScore() + tmp_arc->getWeight(); 	    
	    next_trace->setScore(score);
	    
	    // update the history with the next node at this level
	    //
	    GraphVertex<SearchNode>* target_vertex = tmp_arc->getVertex();
	    next_trace->getHistory()->peek()
	      ->assignAndAdvance((ulong)target_vertex);	    
	    
	    // add the trace to the search node's trace list to make sure we
	    // propagate it before exiting
	    //
	    Trace* existing_trace = (Trace*)NULL;	    
	    SearchNode* next_node =
	      next_trace->getHistory()->peek()->getCentralVertex()->getItem();

	    if (next_node->addTrace(next_trace, current_frame_d,
				    existing_trace)) {
	      
	      // set the backpointer
	      //
	      next_trace->setBackPointer(curr_trace);

	      // add the trace to this level's trace list
	      //
	      trace_lists_d(level_num_a).insertLast(next_trace);
	      if (debug_level_d >= Integral::ALL) {
		printNewPath(next_trace, curr_trace);
	      }
	    }
	    else {

	      // delete the trace since there is existing trace - same history
	      //
	      if (debug_level_d >= Integral::ALL) { 
		printDeletedPath(next_trace, curr_trace);
	      }	  	      
	      delete next_trace;
	    }
	  }
	  
	  // bump the current trace off of the trace list
	  //
	  if (curr_trace->getRefCount() < 1) {
	    Trace::deleteTrace(curr_trace, true);
	  }
	  trace_lists_d(level_num_a).gotoFirst();
	}
      }
    }
  }
  
  // exit gracefully
  //
  return status;
}

// method: beamPruneTrace
//
// arguments:
//  long level_num: (input) the level for which we will propagate traces
//
// return: logical error status.
//
boolean StackSearch::beamPruneTrace(long level_num_a) {

  // define local variables
  //
  Trace* curr_trace = (Trace*)NULL;
  float threshold = max_frame_scores_d(current_frame_d, level_num_a)
    - search_levels_d(level_num_a).getBeamThreshold();
  float curr_score;

  long pruned_traces = 0;
  long orig_length = trace_lists_d(level_num_a).length();
  
  // loop over all active traces at this level
  //
  boolean more_traces = trace_lists_d(level_num_a).gotoFirst();
  
  while (more_traces) {

    curr_trace = trace_lists_d(level_num_a).getCurr();
    curr_score = curr_trace->getScore();

    if (curr_score > max_frame_scores_d(current_frame_d, level_num_a)) {
      return Error::handle(name(), L"beamPrune -- bad maximal score!",
			     Error::ARG, __FILE__, __LINE__);
      }
      
    // prune the traces with too low score
    //
    if (curr_score < threshold)  {

      if (debug_level_d >= Integral::DETAILED) {
	String output(L"score is SMALLER than threshold:");
	output.concat(curr_score);
	output.concat(L" trace:" );
	output.concat(curr_trace);
	Console::put(output);
      }

      if (trace_lists_d(level_num_a).isLast()) {
	more_traces = false;
      }
	  
      // bump the current trace off of the trace list
      //
      trace_lists_d(level_num_a).remove();
      
      if (curr_trace->getRefCount() < 1) {
	  Trace::deleteTrace(curr_trace, true);
      }
      pruned_traces++;
    }
    
    else {
      if (debug_level_d >= Integral::DETAILED) {
	String output(L"score is NOT smaller than threshold:");
	output.concat(curr_score);
	output.concat(L" trace:" );
	output.concat(curr_trace);
	Console::put(output);
      }
      more_traces = trace_lists_d(level_num_a).gotoNext();
    }
    
  }

  if (debug_level_d >= Integral::DETAILED) {

    // print the debugging information
    //
    String output(L"Pf:");
    output.concat(current_frame_d);
    output.concat(L" lv:");
    output.concat(level_num_a);
    output.concat(L" orig_tr:");
    output.concat(orig_length);
    output.concat(L" PR_tr*");
    output.concat(pruned_traces);
    output.concat(L" res_tr:");
    output.concat(trace_lists_d(level_num_a).length());
    output.concat(L" thr:");
    output.concat(threshold);
    output.concat(L" max_sc:");
    output.concat(max_frame_scores_d(current_frame_d, level_num_a));
    output.concat(L"\n");
    Console::put(output);
  }
  
  // exit gracefully
  //
  return true;
}

