// file: $isip/util/speech/isip_model_creator/isip_model_creator.cc
// version: $Id: isip_model_creator.cc,v 1.11 2002/07/16 03:10:00 wang Exp $
//
// isip include files
//
#include "isip_model_creator.h"

// isip_model_creator
// This untility porogram converts four input text files:
// 1. language model file in JSGF
// 2. lexicon file
// 3. acoustic model file in JSGF)
// 4. transcription file
// Then, output three SOF files:
// 1. a language model SOF file containing the converted first three inputs
// 2. a statistical model SOF file containing initialized statistical models
// 3. a transcription SOF file containing a vector of transcription strings
//
// function prototypes
//
boolean getSymbols(String grammar_a, Vector<String>& symbol_list_a);
boolean isSymbol(const String& arg_a);
boolean convertLM(Filename& lm_file_a, Sof& sof_a,
		  Vector<String>& symbol_list_a);
boolean isStart(String& arg_a);
boolean isTerm(String& arg_a);
boolean isNonSpeech(String& arg_a);
boolean convertLexicon(Filename& lexicon_file_a, Sof& sof_a,
		       Vector<String>& input_symbol_list_a,
		       Vector<String>& output_symbol_list_a);
boolean convertAcoustic(Filename& acoustic_file_a, Sof& sof_a,
			Vector<String>& input_symbol_list_a,
			Vector<String>& output_symbol_list_a);
boolean defaultGrammar(String& grammar_a, String& symbol_a,
		       String& def_model_a, String& reserved_word_a,
		       long& reserved_word_counter_a);
boolean getGrammarNames(Vector<String>& grammar_name_list_a,
			Vector<String>& grammar_list_a);
boolean initializeStat(Sof& sof_a, Vector<String>& symbol_list_a,
		       long& num_features_a);
boolean convertTrans(Filename& trans_file_a, Sof& sof_a);

// main application
//
int main(int argc, const char** argv) {
  
  // setup the command line
  //
  Sdb sdb;
  CommandLine cmdl(sdb);  
  cmdl.setUsage(
#include "usage_message.text"
                );
  cmdl.setHelp(
#include "help_message.text"
               );

  cmdl.setIdent("$Revision: 1.11 $",
		"$Name: isip_r00_n11 $",
		"$Date: 2002/07/16 03:10:00 $");

  // add a command line option for the language model file
  //
  Filename lm_file;
  Filename def_lm_input_file(L"lm_model.text");
  cmdl.addOptionParam(lm_file, OPTION_LM_FILE, def_lm_input_file);
  
  // add a command line option for the lexicon file
  //
  Filename lexicon_file;
  Filename def_lexicon_input_file(L"lexicon.text");
  cmdl.addOptionParam(lexicon_file, OPTION_LEXICON_FILE,
		      def_lexicon_input_file);
  
  // add a command line option for the acoustic model file
  //
  Filename acoustic_file;
  Filename def_acoustic_input_file(L"acoustic_model.text");
  cmdl.addOptionParam(acoustic_file, OPTION_ACOUSTIC_FILE,
		      def_acoustic_input_file);

  // add a command line option for the transcription file
  //
  Filename trans_file;
  Filename def_trans_input_file(L"trans.text");
  cmdl.addOptionParam(trans_file, OPTION_TRANSCRIPTION_FILE,
		      def_trans_input_file);

  // add a command line option for the output language model file
  //
  Filename output_lm_file;
  Filename def_lm_output_file(L"lm_model.sof");
  cmdl.addOptionParam(output_lm_file, OPTION_OUTPUT_LM_MODEL,
		      def_lm_output_file);
  
  // add a command line option for the output statistical model file
  //
  Filename output_stat_file;
  Filename def_stat_output_file(L"stat_model.sof");
  cmdl.addOptionParam(output_stat_file, OPTION_OUTPUT_STAT_MODEL,
		      def_stat_output_file);

  // add a command line option for the output transcription file
  //
  Filename output_trans_file;
  Filename def_trans_output_file(L"trans.sof");
  cmdl.addOptionParam(output_trans_file, OPTION_OUTPUT_TRANSCRIPTION,
		      def_trans_output_file);
  
  // parse the command line
  //
  if (!cmdl.parse(argc, argv)) {
    cmdl.printUsage();
  }
  
  // read input model files and convert them into the output LM sof file
  //
  // declare and open output Sof file for storing the converted models
  //
  Sof lm_sof;
  lm_sof.open(output_lm_file, File::WRITE_ONLY, File::TEXT);

  // declare symbol lists
  //
  Vector<String> lm_symbol_list;
  Vector<String> lexicon_symbol_list;
  Vector<String> acoustic_symbol_list;
  
  // read language model text file
  //
  convertLM(lm_file, lm_sof, lm_symbol_list);

  // read lexicon text file
  //
  convertLexicon(lexicon_file, lm_sof, lm_symbol_list, lexicon_symbol_list);
  
  // read accoustic model text file
  //
  convertAcoustic(acoustic_file, lm_sof, lexicon_symbol_list,
		  acoustic_symbol_list);
  
  // close the output sof file
  //
  lm_sof.close();

  // set the number of features
  //
  long num_feature = 1;

  // declare and open output Sof file for storing the statistical models
  //
  Sof stat_sof;
  stat_sof.open(output_stat_file, File::WRITE_ONLY, File::TEXT);

  // initialize statistical models
  //
  initializeStat(stat_sof, acoustic_symbol_list, num_feature);

  // close the output sof file
  //
  stat_sof.close();

  // declare and open output Sof file for storing the transcription
  //
  Sof trans_sof;
  trans_sof.open(output_trans_file, File::WRITE_ONLY, File::TEXT);

  // read the transcription text file
  //
  convertTrans(trans_file, trans_sof);

  // close the output sof file
  //
  trans_sof.close();
  
  // exit gracefully
  //
  return Integral::exit();
}


// function implementations
//
// function: get a list of symbols in a given search level in the sof file
//
boolean getSymbols(String grammar_a, Vector<String>& symbol_list_a) {

  String sub, delim(L" \n;+*");
  long pos = 0;
  boolean rule_reached = false;
  
  // tokenize the input grammar
  //
  while(grammar_a.tokenize(sub, pos, delim)) {

    // search "public" keyword 
    //
    if(sub.eq(L"public")) {
      rule_reached = true;
    }
    // starting from the public rule, get symbols
    //
    if(rule_reached) {
      // if the sub string is a symbol
      //
      if(isSymbol(sub) && sub.ne(L"public")) {
	
	// check if it's already in the symbol list
	//
	boolean exist = false;
	for(long i=0; i<symbol_list_a.length(); i++) {
	  if(sub.eq(symbol_list_a(i))) {
	    exist = true;
	  }
	}
	if(!exist) {
	  // add the symbol to the symbol list
	  //
	  symbol_list_a.concat(sub);
	}
      }
    }
  } // end of the while loop

  // exit gracefully
  //
  return true;
}

// function: check if a string is a symbol
//
boolean isSymbol(const String& arg_a) {

  for(long i=0; i<arg_a.length(); i++) {
    Char tmp = arg_a(i);
    if(!(tmp.isAlpha() || tmp.isDigit() || tmp.eq(L'!') || tmp.eq(L'_')))
      return false;
  }

  // exit gracefully
  //
  return true;
}

// function: convert language model input text file
//
boolean convertLM(Filename& lm_file_a, Sof& sof_a,
		  Vector<String>& symbol_list_a) {

  // number of grammars contained in the input file
  //
  long num_grammars = 0;
  
  // open the input lm_file in read mode
  //
  File read_lm_file;
  if (!read_lm_file.open(lm_file_a, File::READ_ONLY)) {
    Console::put(L"Error in opening language model input file");
  }

  // get the number of grammars in the input file
  //
  String str;
  while(read_lm_file.get(str)) {
    String sub;
    long pos(0);
    str.tokenize(sub, pos);
    if(sub.eq(L"#JSGF"))
      num_grammars++;
  }
  // close the input text file
  //
  read_lm_file.close();

  // check if the number of grammar is correct
  // this LM input file should include one model grammar, two graph
  // default vertices definitions and a grammar for nonspeech symbols.
  // But the vertices definitions and nonspeech definitions are optional
  //
  if(!(num_grammars == 4 || num_grammars == 3 || num_grammars == 1)) {
    return Error::handle(lm_file_a, L"invalid number of grammars\n",
			 Error::TEST, __FILE__, __LINE__);
  }
  
  // declare a vector of strings to contain the input text file
  //
  Vector<String> lm_text(num_grammars);
  
  // open the input file again
  //
  if (!read_lm_file.open(lm_file_a, File::READ_ONLY)) {
    Console::put(L"Error in opening language model input file");
  }
  
  // read the input text and add it line by line to the string vector
  //
  String str_lm, tmp_header;
  long header_counter = 0;

  for(long i=0; i<num_grammars; i++) {

    // add the grammar head line
    //
    lm_text(i).assign(L"grammar = {\n");
    lm_text(i).concat(tmp_header);
      
    // get text line by line
    //
    while(read_lm_file.get(str_lm)) {

      // check if a new grammar is reached
      // (a grammar always starts with #JSGF header)
      //
      String sub;
      long pos(0);
      str_lm.tokenize(sub, pos);
	
      if(sub.eq(L"#JSGF")) {
	if(header_counter>0) {
	  tmp_header.assign(str_lm);
	  tmp_header.concat(L'\n');
	  break;
	}
	header_counter++;
      }
      
      lm_text(i).concat(str_lm);
      lm_text(i).concat(L'\n');
    }

    // add the grammar end line
    //
    lm_text(i).concat(L"};\n\n");
  } // end of for loop
  
  // close the input lm_file
  //
  read_lm_file.close();

  
  // start storing the input language models in the input sof file
  //
  // declare strings to contain the grammars for the default
  // ISIP graph vertices, nonspeech grammars, and lm grammars
  //
  String start_def, term_def, nonsp_def, lm_grammar;
  
  // declare the line of algorithm tag
  //
  String algo_tag(L"algorithm = \"JSGF\";\n");

  // set the size of the object to be written
  //
  long size_1 = algo_tag.length();
  long size_2 = algo_tag.length();

  // boolean variables to indicate if there are definitions
  // for graph ending points
  //
  boolean def_start = false;
  boolean def_term = false;
  boolean def_nonsp = false;
  
  // loop throught all the grammars in the input LM text file
  //
  for (long i=0; i<lm_text.length(); i++) {

    if (isStart(lm_text(i))) {
      start_def.concat(lm_text(i));
      def_start = true;
    }
    else if (isTerm(lm_text(i))) {
      term_def.concat(lm_text(i));
      def_term = true;
    }
    else if (isNonSpeech(lm_text(i))) {
      nonsp_def.concat(lm_text(i));
      def_nonsp = true;
    }
    else {
      // this is LM grammar
      //
      lm_grammar.concat(lm_text(i));
    }
  }

  // set size
  //
  size_1 += start_def.length() + term_def.length();
  size_2 += nonsp_def.length() + lm_grammar.length();
  
  // write the definitions of graph starting and stoping points
  // into the sof file
  //
  if (def_start && def_term) {
    sof_a.put(L"JSGF", 100, size_1);
    sof_a.puts(algo_tag);
    sof_a.puts(start_def);
    sof_a.puts(term_def);
  }
  
  // write the language model into the sof file
  //
  sof_a.put(L"JSGF", 0, size_2);
  sof_a.puts(algo_tag);
  if (def_nonsp) {
    sof_a.puts(nonsp_def);
  }
  sof_a.puts(lm_grammar);

  // get the symbol list in this level
  //
  getSymbols(lm_grammar, symbol_list_a);
  
  // exit gracefully
  //
  return true;    
}

// function: check if the input grammar string is a definition for
// ISIP default graph starting point
//
boolean isStart(String& arg_a) {

  String sub, input(arg_a), delim(L" \n");
  long pos = 0;

  // search "public" keyword
  //
  while(input.tokenize(sub, pos, delim)) {
    if(sub.eq(L"public")) {

      // check the next token
      //
      input.tokenize(sub, pos, delim);      
      if(sub.eq(L"<ISIP_JSGF_1_0_START>")) {

	// found the start symbol definition
	//
	return true;
      }
    }
  }
  // start symbol definition not found
  //
  return false;
}

// function: check if the input grammar string is a definition for
// ISIP default graph terminal point
//
boolean isTerm(String& arg_a) {

  String sub, input(arg_a), delim(L" \n");
  long pos = 0;

  // search "public" keyword
  //
  while(input.tokenize(sub, pos, delim)) {
    if(sub.eq(L"public")) {

      // check the next token
      //
      input.tokenize(sub, pos, delim);      
      if(sub.eq(L"<ISIP_JSGF_1_0_TERM>")) {

	// found the terminal symbol definition
	//
	return true;
      }
    }
  }
  // terminal symbol definition not found
  //
  return false;
}

// function: check if the input grammar string is a definition for
// ISIP reserved nonspeech symbols
//
boolean isNonSpeech(String& arg_a) {

  String sub, input(arg_a), delim(L" \n");
  long pos = 0;

  // search "public" keyword
  //
  while(input.tokenize(sub, pos, delim)) {
    if(sub.eq(L"public")) {

      // check the next token
      //
      input.tokenize(sub, pos, delim);      
      if(sub.eq(L"<ISIP_NON_SPEECH>")) {
	return true;
      }
    }
  }
  // the definition not found
  //
  return false;
}

// function: convert lexicon input text file
//
boolean convertLexicon(Filename& lexicon_file_a, Sof& sof_a,
		       Vector<String>& input_symbol_list_a,
		       Vector<String>& output_symbol_list_a) {

  // number of grammars contained in the input file
  //
  long num_grammars = 0;
  boolean def_nonsp = false;
  
  // open the input file in read mode
  //
  File read_lexicon_file;
  if (!read_lexicon_file.open(lexicon_file_a, File::READ_ONLY)) {
    Console::put(L"Error in opening lexicon input file");
  }

  // declare variables
  //
  String str;
  Vector<String> nonsp_def, pre_list, word_list, rule_list;
  boolean JSGF_found = false;
  
  // read each line
  //
  while(read_lexicon_file.get(str)) {

    // look for the JSGF grammar that defines the nonspeech symbols
    //
    if (str.eq(L"#JSGF V1.0;")) {
      JSGF_found = true;
    }

    // after meeting the JSGF grammar, get the JSGF grammar that
    // defines the nonspeech symbols
    //
    if (JSGF_found) {
	nonsp_def.concat(str);
    }

    // before meeting the JSGF grammar, process lexicon
    //
    else {
      // pre-process the input lexicon lines and merge the same lines
      //
      boolean same = false;
      for(long i=0; i<pre_list.length(); i++) {
	if(str.eq(pre_list(i))) 
	  same = true;
      }
      if(!same) { 
	pre_list.concat(str);
      }
    }
  }

  // create grammar line for the nonspeech JSGF grammar
  //
  String nonsp_grammar(L"grammar = {\n");
  for (long i = 0; i < nonsp_def.length(); i++) {
    nonsp_grammar.concat(L"  ");
    nonsp_grammar.concat(nonsp_def(i));
    nonsp_grammar.concat(L"\n");
  }
  nonsp_grammar.concat(L"};\n\n");
  
  // check if non-speech grammar is available
  //
  if (isNonSpeech(nonsp_grammar)) {
    def_nonsp = true;
  }
    
  // process each lexicon line after pre-processing in the pre_list
  //
  for(long i=0; i<pre_list.length(); i++) {
    
    String head_word, symbol, sequence;
    long pos(0);
    String delim(L" ");
    String lex_str(pre_list(i));

    // get the first word in the lexicon line
    //
    lex_str.tokenize(head_word, pos);

    // tokenize out each symbol sequentially from the lexicon line
    //
    while(lex_str.tokenize(symbol, pos, delim)) {
      if(symbol.ne(head_word)) {
	sequence.concat(L" /0/ ");
	sequence.concat(symbol);
      }
    }

    // check if the head_word already exists in the word list
    //
    boolean exist = false;
    long alt_index = 0;
    
    for(long i=0; i<word_list.length(); i++) {
      if(head_word.eq(word_list(i))) {
	exist = true;
	alt_index = i;
      }
    }

    // if the head word is not available in the word_list yet 
    //
    if(!exist) {
      // add the head word to the word_list
      //
      word_list.concat(head_word);
      num_grammars++;

      // also add the rule sequence to the rule_list
      //
      rule_list.concat(sequence);
    }
    
    // if the head word is already included in the word_list
    //
    else {
      // add the sequence rule part to the rule list as alternative
      //
      String tmp;

      // get existing rule in the rule line
      //
      String exist_rule(rule_list(alt_index));
      Char end(exist_rule(rule_list(alt_index).length()-1));

      // if no alternative relation exists in the rule line
      // ( and ) signs need to be added to surround the existing sequence rule
      // so that we add new alternative parallel to the existing sequence
      //
      if(!end.eq(')')) {
	tmp.assign(L" (");
	tmp.concat(exist_rule);
	tmp.concat(L" )");
      }
      else {
	tmp.assign(exist_rule);
      }

      // add the new sequence as alternative
      //
      tmp.concat(L" | (");
      tmp.concat(sequence);
      tmp.concat(L" )");

      // update the rule in the rule list
      //
      rule_list(alt_index).assign(tmp);
    }
  } // end: for(long i=0; i<pre_list.length(); i++) 
  
  // close the input text file
  //
  read_lexicon_file.close();
  
  // make sure each symbol in the input symbol list has a corresponding
  // grammar defined in the lexicon file
  //
  if(num_grammars != input_symbol_list_a.length()) {
    // output warning
    //
    Console::put(L"\nWarning: the number of grammars in the lexicon input file\ndoes not match the number of symbols in the LM input file.\n\n");
  }
  
  Vector<String> words, rules;
  for(long i=0; i<input_symbol_list_a.length(); i++) {

    boolean found = false;
    for(long j=0; j<word_list.length(); j++) {

      if(input_symbol_list_a(i).eq(word_list(j))) {
	found = true;
	words.concat(word_list(j));
	rules.concat(rule_list(j));
      }
    }
    if(!found) {
      String output(L"symbol \"");
      output.concat(input_symbol_list_a(i));
      output.concat(L"\"");
      return Error::handle(output, L"no corresponding grammar in lexicon file\n", Error::TEST, __FILE__, __LINE__);
    }
  }

  // update the number of lexicon grammars to fit the input symbol list
  //
  num_grammars = words.length();
  
  // declare a vector of string to store JSGF grammar strings
  //
  Vector<String> grammars(num_grammars);

  // declare and initialize a string to store a single JSGF grammar
  //
  String grammar_str(L"grammar = {\n");
  grammar_str.concat(L"  #JSGF V1.0;\n");
  grammar_str.concat(L"  // Define the grammar name\n");
  grammar_str.concat(L"  grammar network.grammar.");

  for(long i=0; i<grammars.length(); i++) {

    grammars(i).assign(grammar_str);
    grammars(i).concat(words(i));
    grammars(i).concat(L";\n\n  // Define the rules\n  public <");
    grammars(i).concat(words(i));
    grammars(i).concat(L"> = ");
    grammars(i).concat(L"<ISIP_JSGF_1_0_START>");

    Char end(rules(i)(rules(i).length()-1));
    
    // if the rule includes alternative
    //
    if(end.eq(L')')) {
      grammars(i).concat(L" (");
      grammars(i).concat(rules(i));
      grammars(i).concat(L" )");
    }
    // if it includes only sequence
    //
    else {
      grammars(i).concat(rules(i));
    }

    grammars(i).concat(L" /0/ <ISIP_JSGF_1_0_TERM>;\n");
    grammars(i).concat(L"};\n\n");
  }

  // write the grammars into the output Sof file
  //
  // define algorithm tag line
  //
  String algo_tag(L"algorithm = \"JSGF\";\n");

  // set the size of the object to be written
  //
  long size = algo_tag.length();
  
  if (def_nonsp) {
    size += nonsp_grammar.length();
  } 
  
  for(long i=0; i<grammars.length(); i++) {
    size += grammars(i).length();
  }
    
  // write the lexicon model into the sof file
  //
  sof_a.put(L"JSGF", 1, size);
  sof_a.puts(algo_tag);

  if (def_nonsp) {
    sof_a.puts(nonsp_grammar);
  }
  
  for(long i=0; i<grammars.length(); i++) {
    sof_a.puts(grammars(i));
  }

  // get the symbol list in this level
  //
  for(long i=0; i<grammars.length(); i++) {

    Vector<String> tmp, symbol_list;
    getSymbols(grammars(i), tmp);

    // check if any symbol is already in the symbol list
    //
    for(long i=0; i<tmp.length(); i++) {
      boolean exist = false;
      for(long j=0; j<output_symbol_list_a.length(); j++) {
 
	if(tmp(i).eq(output_symbol_list_a(j))) {
	  exist = true;
	}
      }
      if(!exist) {
	symbol_list.concat(tmp(i));
      }
    }
    output_symbol_list_a.concat(symbol_list);
  }
  
  // exit gracefully
  //
  return true;
}

// function: convert acoustic model input text file
//
boolean convertAcoustic(Filename& acoustic_file_a, Sof& sof_a,
			Vector<String>& input_symbol_list_a,
			Vector<String>& output_symbol_list_a) {

  // number of grammars contained in the input file
  //
  long num_grammars = 0;
  
  // open the input file in read mode
  //
  File read_acoustic_file;
  if (!read_acoustic_file.open(acoustic_file_a, File::READ_ONLY)) {
    Console::put(L"Error in opening acoustic model input file");
  }

  // get the number of grammars in the input file
  //
  String str;
  while(read_acoustic_file.get(str)) {
    String sub;
    long pos(0);
    str.tokenize(sub, pos);
    if(sub.eq(L"#JSGF"))
      num_grammars++;
  }
  // close the input text file
  //
  read_acoustic_file.close();
  
  // declare a vector of strings to contain the input text file
  //
  Vector<String> acoustic_text(num_grammars);
  
  // open the input file again
  //
  if (!read_acoustic_file.open(acoustic_file_a, File::READ_ONLY)) {
    Console::put(L"Error in opening acoustic model input file");
  }
  
  // read the input text and add it line by line to the string vector
  //
  String str_lm, tmp_header;
  long header_counter = 0;

  for(long i=0; i<num_grammars; i++) {

    // add the grammar head line
    //
    acoustic_text(i).assign(L"grammar = {\n ");
    acoustic_text(i).concat(tmp_header);

    // get text line by line
    //
    while(read_acoustic_file.get(str_lm)) {

      // check if a new grammar is reached
      // (a grammar always starts with #JSGF header)
      //
      String sub;
      long pos(0);
      str_lm.tokenize(sub, pos);
	
      if(sub.eq(L"#JSGF")) {
	if(header_counter>0) {
	  tmp_header.assign(str_lm);
	  tmp_header.concat(L'\n');
	  break;
	}
	header_counter++;
      }

      acoustic_text(i).concat(str_lm);
      acoustic_text(i).concat(L'\n');
    }

    // add the grammar end line
    //
    acoustic_text(i).concat(L"};\n\n");
  } // end of for loop
  
  // close the input acoustic file
  //
  read_acoustic_file.close();

  // get grammar names from the text grammars
  //
  Vector<String> grammar_name_list;
  getGrammarNames(grammar_name_list, acoustic_text);
  
  // useful variables
  //
  boolean default_model = false;
  boolean reserved_symbol = false;
  long def_model_index = 0;
  long reserved_word_index = 0;
  String reserved_word;

  // loop through the grammar name list for searching
  //
  for (long i = 0; i < grammar_name_list.length(); i++) {

    // 1: search the grammar defining ISIP default acoustic model
    // in the input grammars
    //
    if (grammar_name_list(i).eq(L"ISIP_JSGF_1_0_DEFAULT_ACOUSTIC_MODEL")) {
      default_model = true;
      def_model_index = i;
    }
  
    // 2: search the grammar defining ISIP user-defined reserved symbol
    // in the input grammars
    //
    if (grammar_name_list(i).eq(L"USER_RESERVED_SYMBOL")) {
      reserved_symbol = true;
      reserved_word_index = i;

      // get the reserved symbol
      //
      String sub, delim(L" };\n");
      long pos = 0;
         
      // tokenize the grammar text
      //
      while(acoustic_text(i).tokenize(sub, pos, delim)) {
	// search "public" keyword 
	//
	if(sub.eq(L"public")) {
	  // get the token after "=" sign in the public rule line
	  //
	  acoustic_text(i).tokenize(sub, pos, delim);
	  acoustic_text(i).tokenize(sub, pos, delim);
	  acoustic_text(i).tokenize(sub, pos, delim);

	  // the token is the user-defined reserved symbol
	  //
	  reserved_word.assign(sub);
	}
      } // end of the while loop 
    } // end: if (grammar_name_list(i).eq(L"USER_RESERVED_SYMBOL"))
  }

  // make sure the reserved symbol is not used in the grammars
  // other than the default model grammar and the reservation grammar
  //
  if(reserved_symbol) {
    // check through all input grammars
    //
    for(long i = 0; i<acoustic_text.length(); i++) {
      Vector<String> tmp;

      // skip the two reserved default definition grammars
      //
      if((i != reserved_word_index) && (i != def_model_index)) {

	// get symbol list in the given grammar
	//
	getSymbols(acoustic_text(i), tmp);
	
	// check if the reserved word is used by each symbol
	//
	for(long j = 0; j < tmp.length(); j++) {
	  String sub_str, delim(L" _0123456789");
	  long pos = 0;
	  tmp(j).tokenize(sub_str, pos, delim);

	  // if any grammar illegally uses the reserved word
	  // output the grammar and return error message
	  //
	  if(sub_str.eq(reserved_word)) {

	    String output_grammar(acoustic_text(i));
	    Console::put(L"\n\nThe following grammar is illegal in using reserved symbol\n\n");
	    Console::put(output_grammar);
	    
	    return Error::handle(L"acoustic input file", L"the user-reserved symbol has been illegally used in other grammar definitions\n", Error::TEST, __FILE__, __LINE__);
	  }
	} // end: for(long j = 0; j < tmp.length(); j++) {
      } // end: if((i != reserved_word_index) && (i != def_model_index))
    } // end: for(long i = 0; i<acoustic_text.length(); i++)
  }
 
  // make sure each input symbol in the upper level has a corresponding
  // grammar in this level
  //
  Vector<String> grammars;
  long reserved_word_counter = 1;
  
  for (long i = 0; i < input_symbol_list_a.length(); i++) {

    boolean found = false;
    for(long j=0; j<grammar_name_list.length(); j++) {

      if(input_symbol_list_a(i).eq(grammar_name_list(j)) ||
	 grammar_name_list(j).eq(L"ISIP_JSGF_1_0_START") ||
	 grammar_name_list(j).eq(L"ISIP_JSGF_1_0_TERM")) {
	
	found = true;
	grammars.concat(acoustic_text(j));
      }
    }
    
    if(!found) {
      // if no corresponding grammar defined, use the user-defined
      // default acoustic model grammar
      //
      if(default_model) {
	String message(L"Note: symbol \"");
	message.concat(input_symbol_list_a(i));
	message.concat(L"\" uses the user-defined default acoustic model.\n");
	Console::put(message);

	// create a grammar for the given symbol according to the default
	//
	String tmp_grammar;

	if(reserved_symbol) {
	  defaultGrammar(tmp_grammar, input_symbol_list_a(i),
			 acoustic_text(def_model_index), reserved_word,
			 reserved_word_counter);
	}
	else {
	  // if user does not define the reserved symbol word
	  //
	  return Error::handle(L"substitution for default acoustic model", L"user has not yet defined the reserved symbol for the substitution.\nPlease add a JSGF grammar to define the reserved symbol in the input acoustic text file!\n", Error::TEST, __FILE__, __LINE__);
	}

	grammars.concat(tmp_grammar);
      }
      
      // if neither corresponding grammar nor default model grammar
      // can be found, return error message
      //
      else {
	String output(L"symbol \"");
	output.concat(input_symbol_list_a(i));
	output.concat(L"\"");
	return Error::handle(output, L"neither corresponding grammar nor default model defined in the acoustic file\n", Error::TEST, __FILE__, __LINE__);
      }
    }
  }
 
  // then store the input acoustic models in the input sof file
  //
  // declare strings to contain the grammars of definitions for
  // ISIP graph starting and stoping points
  //
  String start_def, term_def;

  // vector of string to store acoustic grammars
  //
  Vector<String> acoustic_grammar;
 
  // declare the line of algorithm tag
  //
  String algo_tag(L"algorithm = \"JSGF\";\n");

  // set the size of the object to be written
  //
  long size_1 = algo_tag.length();
  long size_2 = algo_tag.length();

  // boolean variables to indicate if there are definitions
  // for graph ending points
  //
  boolean def_start = false;
  boolean def_term = false;
  
  // loop throught all the grammars in the input acoustic text file
  //
  for(long i=0; i < grammars.length(); i++) {

    if(isStart(grammars(i))) {
      start_def.assign(grammars(i));
      size_1 += grammars(i).length();
      def_start = true;
    }
    else if(isTerm(grammars(i))) {
      term_def.assign(grammars(i));
      size_1 += grammars(i).length();
      def_term = true;
    }
    else {
      // this is acoustic grammar
      //
      acoustic_grammar.concat(grammars(i));
      size_2 += grammars(i).length();
    }
  }

  // write the definitions of graph starting and stoping points
  // into the sof file, if they are available
  //
  if(def_start && def_term) {

    sof_a.put(L"JSGF", 100, size_1);
    sof_a.puts(algo_tag);
    sof_a.puts(start_def);
    sof_a.puts(term_def);
  }
  
  // write the acoustic model into the sof file
  //
  sof_a.put(L"JSGF", 2, size_2);
  sof_a.puts(algo_tag);
  for(long i=0; i<acoustic_grammar.length(); i++) {
    sof_a.puts(acoustic_grammar(i));
  }


  // get the symbol list in this level
  //
  for(long i=0; i<acoustic_grammar.length(); i++) {

    Vector<String> tmp, symbol_list;
    getSymbols(acoustic_grammar(i), tmp);
    
    // check if any symbol is already in the symbol list
    //
    for(long i=0; i<tmp.length(); i++) {
      boolean exist = false;
      for(long j=0; j<output_symbol_list_a.length(); j++) {
 
	if(tmp(i).eq(output_symbol_list_a(j))) {
	  exist = true;
	}
      }
      if(!exist) {
	symbol_list.concat(tmp(i));
      }
    }
    output_symbol_list_a.concat(symbol_list);
  }
    
  // exit gracefully
  //
  return true;  
}

// function: create an acoustic grammar based on the user-defined
// default acoustic model grammar
//
boolean defaultGrammar(String& grammar_a, String& symbol_a,
		       String& def_model_a, String& reserved_word_a,
		       long& reserved_word_counter_a) {

  // process the rule definition in the default model
  //
  String rule;
  String sub, delim(L" };\n");
  long pos = 0;
  boolean rule_reached = false;

  // preprocess the input model string by picking up the actual model grammar
  // surrounded by { and }
  //
  String def_model;
  def_model_a.substr(def_model, 12, def_model_a.length() - 12 - 4);
  
  // then tokenize the actual model grammar string
  //
  delim.assign(L" ;\n");
  while(def_model.tokenize(sub, pos, delim)) {

    // search "public" keyword 
    //
    if(sub.eq(L"public")) {
      // get to the token after "=" sign in the public rule line
      //
      def_model.tokenize(sub, pos, delim);
      def_model.tokenize(sub, pos, delim);
      
      rule_reached = true;
    }

    // process following symbols
    //
    if(rule_reached) {

      delim.assign(L" \n");      
      while(def_model.tokenize(sub, pos, delim)) {

	// add space between each two adjacent tokens
	//
	rule.concat(L" ");
  
	// discard ";", "+" or "*" sign for further symbol check
	//
	String tmp;
	Char end(sub(sub.length() - 1));
	boolean discard = false;
	boolean semicolon = false;
	
	// if a semi-colon is found at the end of the symbol
	//
	if (end.eq(L';')) {
	  sub.substr(tmp, 0, sub.length() - 1);
	  sub.assign(tmp);
	  end.assign(sub(sub.length() - 1));
	  semicolon = true;
	}

	// if "*" or "+" is found at the end of the symbol
	//
	if(end.eq(L'*') || end.eq(L'+')) {
	  sub.substr(tmp, 0, sub.length() - 1);
	  sub.assign(tmp);
	  discard = true;
	}
	
	// if a symbol is met, substitute it as the default symbol
	//
	if(isSymbol(sub)) {
	  sub.assign(reserved_word_a);
	  sub.concat(L"_");
	  sub.concat(reserved_word_counter_a);
	  reserved_word_counter_a++;
	}

	// if discarding already happened, add the discarded
	// "*" or "+" sign back to the sub-string
	//
	if(discard) {
	  sub.concat(end);
	}

	// if semicolon has been met and discarded, add the discarded
	// ";" back to tmp with an additional new-line character
	//
	if(semicolon) {
	  sub.concat(L";\n ");
	}
	
	// add the rule component to the new rule line
	//
	rule.concat(sub);
      }
    }
  } // end of the outer while loop

  // drop the space at the end of the rule string
  //
  String tmp_rule;
  rule.substr(tmp_rule, 0, rule.length() - 1);
  rule.assign(tmp_rule);
  
  // create a substituted grammar for the input symbol
  //
  grammar_a.assign(L"grammar = {\n  #JSGF V1.0;\n");
  grammar_a.concat(L"  // Define the grammar name\n");
  grammar_a.concat(L"  grammar network.grammar.");
  grammar_a.concat(symbol_a);
  grammar_a.concat(L";\n\n");
  grammar_a.concat(L"  // Define the rules using the default model\n");
  grammar_a.concat(L"  public <");
  grammar_a.concat(symbol_a);
  grammar_a.concat(L"> =");
  grammar_a.concat(rule);
  grammar_a.concat(L"};\n\n");
  
  // exit gracefully
  //
  return true; 
}

// function: get grammar names from given grammars
//
boolean getGrammarNames(Vector<String>& grammar_name_list_a,
			Vector<String>& grammar_list_a) {
  
  for(long i=0; i<grammar_list_a.length(); i++) {

    String sub, delim(L" \n;");
    long pos = 0;
    String tmp(grammar_list_a(i));
    // tokenize the input grammar
    //
    while(tmp.tokenize(sub, pos, delim)) {
      // search "public" keyword 
      //
      if(sub.eq(L"public")) {
	// get the folowing public rulename
	//
	tmp.tokenize(sub, pos, delim);

	// take off the surrounding < and > signs
	//
	String grammar_name;
	sub.substr(grammar_name, 1, sub.length()-2);

	// add it to the list
	//
        grammar_name_list_a.concat(grammar_name);
      }
    } // end of the while loop
  } // end of the for loop
  
  // exit gracefully
  //
  return true;
}

// function: initialize statistical models based on given symbols
//
boolean initializeStat(Sof& sof_a, Vector<String>& symbol_list_a,
		       long& num_features_a) {

  // number of symbols in the input symbol list
  //
  long num_symbols = 0;
  
  // declare a hash table
  //
  HashTable<SearchSymbol, Long> symbol_hash;

  // insert each symbol and its index into the hash table
  //
  for(long i = 0; i < symbol_list_a.length(); i++) {

    SearchSymbol symbol;
    Long* index;

    symbol.assign(symbol_list_a(i));
    index = new Long(i);
    symbol_hash.insert(symbol, index);

    // get the number of symbols
    //
    num_symbols++;
  }

  // write the hash table into the output sof file
  //
  symbol_hash.write(sof_a, 2, L"symbol_hashtable");    


  // initialize statistical models
  //
  // declare variables
  //
  Vector<MixtureModel> mix_models(num_symbols);
  Vector<StatisticalModel> stat_models(num_symbols);
  VectorFloat weights(1), mean(num_features_a);
  MatrixFloat cov;
  String mean_str, value_str;

  // set Gaussian model
  // create mean and value strings
  //
  for (long i = 0; i < num_features_a - 1; i++) {
    mean_str.concat(L"0.0, ");
    value_str.concat(L"1.0, ");
  }
  mean_str.concat(L"0");
  value_str.concat(L"1.0");
    
  // set mean
  //
  mean.assign(mean_str);
  
  // set covariance
  //
  cov.assign(num_features_a, num_features_a, value_str, Integral::DIAGONAL);

  // set Gaussian
  //
  GaussianModel gau_model;
  gau_model.setMean(mean);
  gau_model.setCovariance(cov);
    
  // set weights for mixture model
  //
  weights.assign(L"1");
  
  // set each stat models
  //
  for (long i = 0; i < num_symbols; i++) {  

    // set mixture model
    //
    mix_models(i).setWeights(weights);
    mix_models(i).add(gau_model);

    // set statistical model
    //
    stat_models(i).setType(StatisticalModel::MIXTURE_MODEL);
    stat_models(i).assign(mix_models(i));
  }
  
  // write the statistical models into the output sof file
  //
  stat_models.write(sof_a, 2, L"stat_models");    
  
  // exit gracefully
  //
  return true;
}

// function: read transcription input text file and output it into a sof file
//
boolean convertTrans(Filename& trans_file_a, Sof& sof_a) {

  // declare a string vector to store the transcription information
  //
  Vector<String> trans;
  
  // open the input file in read mode
  //
  File read_trans_file;
  if (!read_trans_file.open(trans_file_a, File::READ_ONLY)) {
    Console::put(L"Error in opening transcription input file");
  }

  // read the string lines
  //
  String input_line;
  while(read_trans_file.get(input_line)) {
    trans.concat(input_line);
  }
  
  // close the input text file
  //
  read_trans_file.close();

  // write the transcription into the output sof file
  //
  trans.write(sof_a, 0);

  // exit gracefully
  //
  return true;
}
